#include <tools.H>
#include <lob.H>
#include <forwarder.H>
#include <tools.H>
#include <Python.h>
#include "anep.H"
#include "pad.H"
#include "pybar.H"

//
// the starter macro
//

LOBSTART(Pybar);


//
// the pybar C O N S T R U C T O R
//

Pybar::Pybar(LOBenv *e) : LOB(e)
{
  debug=0xffff;
  char debug_value[16];
  
  pybar_core_code="core.py";

  iqueue_len=64;
  packets_in_iqueue=0;
  iq_idx_out=0;
  iq_idx_in=0;

  if(arg_query(env->argc, env->argv,'?',NULL,0) ||
     arg_query(env->argc, env->argv,'h',NULL,0)){
    fhprintf(*env->cfh,"pybar [-h] [-?] [-d]\n");
    return;
  }
  
  if(arg_query(env->argc, env->argv, 'd',debug_value,15))debug=atoi(debug_value);
  
  if(debug&PYBAR_INITEXIT){
    fhprintf(*env->cfh,"pybar_init: pybar started (debug mode)\n");
  }

  env->set_mode(LOB_MODE_THREAD);

  // Initialise the packet queue

  iqueue=(Packet**)malloc(sizeof(Packet*)*iqueue_len);
  memset(iqueue,0,sizeof(Packet*)*iqueue_len);

  // Filter Setup
  // all packets for the router alert option shall be
  // forwarded to the pybar system

  struct FILTER_ADDDEL_CB filter_add_cb;
  
  filter_add_cb.filter.source_ip=0;
  filter_add_cb.filter.source_nm=0;
  filter_add_cb.filter.dest_ip=0;
  filter_add_cb.filter.dest_nm=0;
  
  filter_add_cb.handle=0;
  filter_add_cb.len=sizeof(struct FILTER_ADDDEL_CB);
  filter_add_cb.id=30;                     // filter specific command
  filter_add_cb.filter_command=2;          // setup filter
  filter_add_cb.filter.tos=0;
  filter_add_cb.filter.prio=10;
  filter_add_cb.filter.mode=0; 
  filter_add_cb.filter.forwarder=this; 

  strncpy(filter_add_cb.filter.name,env->name,31);
  filter_add_cb.filter.name[31]=0;

  filter_add_cb.filter.proto=0;
  filter_add_cb.filter.ipopt_type=IPOPT_RA;
  filter_add_cb.filter.ipopt_value=0;
  
  fhprintf(*env->cfh, "pybar_init: setting up filter\n");

  struct dummy : public VRRB{
    struct FILTER_REC filter;
  } *rb;
  
#ifdef DEBUG_API
  fhprintf(*env->cfh, "DEBUG_API: pybar setting up filter\n");
#endif

  (VRRB*)rb=env->api_direct((VRCB*)&filter_add_cb);
  
  if(!rb){
    fhprintf(*env->cfh, "pybar_init: error during filter set up\n");
    return;
  }

  if(debug&PYBAR_INITEXIT){
    fhprintf(*env->cfh,"pybar_init: python core goes online\n");
  }

  // start the pybar

  Py_SetProgramName("pybar");
  Py_Initialize();

  if(debug&PYBAR_INITEXIT){
    PyRun_SimpleString("print \"pybar_init: I'm alive\"");
  }

  // Kick up the Platform Adaptor

  init_pad(this);

  // run the Python core

  return;
}

// This is calle dby the VR loader to 
// shoot up the thread with the pybat engine
// the pybar engine itself is written in python

int Pybar::run()
{
  if(debug&PYBAR_INITEXIT){
    fhprintf(*env->cfh,"run() starts\n");
  }

  FILE* fp=fopen(pybar_core_code,"r");
  PyRun_SimpleFile(fp, pybar_core_code);
  fclose(fp);

  if(debug&PYBAR_INITEXIT){
    fhprintf(*env->cfh,"run() ends\n");
  }

  return LOB_TERMINATE;
}


// the forwarding function
// it's called when a new packet arrives for the pybar
// the packet is simply put in the pybar's input queue

int Pybar::forward(Packet *p)
{
  // if too many packets ignore it
  
  if(packets_in_iqueue>=iqueue_len){
    if(debug&PYBAR_PACKET_RECEPTION){
      fhprintf(*env->cfh,"pybar_debug: queue full, discaring packet !!\n");
    }
    return FFILTER_MODE_COPY;
  }

  // enque the packet
  
  iqueue[iq_idx_in]=(Packet *)malloc(PKT_LENGTH(p));
  memcpy(iqueue[iq_idx_in],p,PKT_LENGTH(p));
  iq_idx_in++;
  packets_in_iqueue++;
  if(iq_idx_in>=iqueue_len)iq_idx_in=0; // wrap
  
  return FFILTER_MODE_COPY;
}


//
// the pybar D E S T R U C T O R
//

Pybar::~Pybar() 
{
  if(debug&PYBAR_INITEXIT){
    fhprintf(*env->cfh,"Killing the Interpreters\n");
  }
  
  //
  // empty the packet queue
 
  if(iqueue){
    packets_in_iqueue=0;
    for(unsigned int i=0;i<iqueue_len;++i)if(iqueue[i])free(iqueue[i]);
    free(iqueue);
  }
  
  Py_Finalize();

  if(debug&PYBAR_INITEXIT){
    fhprintf(*env->cfh,"Destructor End\n");
  }
  return;
}






