#include <stdlib.h>
#include <string.h>
#include "config.H"
#include "vr.H"
#include "api.H"
#include "forwarder.H"
#include "api_interface.H"
#include "api_routing.H"
#include "packet.H"
#include "ffilter.H"
#include "interface.H"
#include "routing.H"
#include "tools.H"
#include "vrevent.H"

API::API(VR *v,int mode)
{
  vr=v;
  cb=0;
  api_modus=0;

#ifdef DEBUG_API
  fhputs("API: API constructor",vr->cfh);
#endif

  if(mode){
    api_modus=1;
#ifdef DEBUG_API
    fhputs("API: API constructor active mode",vr->cfh);
#endif
    pipe(api_pipe);
    pipe(api_msg);
    event = new VRevent(this);
    event->type=IO_READ;		// IO_WRITE;
    event->suspended=0;
    event->suspend_time=HUGE_VAL;
    event->thrown=0;
    event->fh_r=api_pipe[0];	// fh_r;
    event->fh_w=0 ;		// fh_w;
    vr->vrsched->notify(event);
    cb=(VRCB*)malloc(0xffff);
    idstring="api";
  }

  handle=42;
  
  return;
}

/*
 * opens an API on two exisiting File handles
 * Control blocks are read from fh_r
 * Result blocks are wirrten to fh_w
 */

API::API(VR *v, int fh_r, int fh_w)
{
  vr=v;
  cb=0;
  api_modus=0;
  
#ifdef DEBUG_API
  fhprintf(vr->cfh,"API: new API generated from file handles %d %d\n", 
	   fh_r, fh_w);
#endif

  idstring="api";
  event = new VRevent(this);
  event->type=IO_READ;		// IO_WRITE;
  event->suspended=0;
  event->suspend_time=HUGE_VAL;
  event->thrown=0;
  event->fh_r=fh_r;
  event->fh_w=fh_w;
  vr->vrsched->notify(event);
  cb=(VRCB*)malloc(0xffff);
  api_pipe[0]=fh_r;
  api_msg[1]=fh_w;
  return;
}



API::~API()
{
  if(cb)free(cb);
  return;
}

int API::write_CB(VRCB *block)
{
  if(!api_modus)return 0;
  return write(api_pipe[1],block,block->len);
}

int API::read_CB(VRRB *block, int l)
{
  if(!api_modus)return 0;
  return read(api_msg[0], block, l);
}


int API::interrupt(VRevent *e)
{
  if(e->thrown&IO_READ){
    int len;

    len=read(api_pipe[0], cb, sizeof(VRCB));

    if(len!=sizeof(VRCB)){
#ifdef DEBUG_API
      fhputs("API: garbled Control Block, stage 1",vr->cfh);
#endif
      return 0;
    }

#ifdef DEBUG_API
    fhprintf(vr->cfh,"API: got VRCB: length %u id %u handle %u\n",cb->len,cb->id,cb->handle);
#endif

    len=read(api_pipe[0], cb+1, cb->len-sizeof(VRCB));
    
    if((unsigned)len != cb->len-sizeof(VRCB)){
#ifdef DEBUG_API
      fhputs("API: garbled Control Block, stage 2",vr->cfh);
#endif
      return 0;
    }

    //
    // THE FOLLOWING FUNCTION HAS ONLY TO BE CALLED WHEN THE BASE SYSTEM IS  
    // STOPPED. SO CALL IT ONLY FROM A SCHEDULED OBJECT !!!
    //
    
    VRRB *rb=process_VRCB(cb);

#ifdef DEBUG_API
    fhprintf(vr->cfh,"API: sending RB with: length %u  handle %u\n", rb->len, rb->handle);
#endif
    write(api_msg[1],rb, rb->len);
    return 0;
  }
  return 0;
}


VRRB *API::process_VRCB(VRCB *cb)
{

#ifdef LOG_APICOM
  char buffer[256];
  sprintf(buffer,"%s%d.cb",LOG_APICOM,cb->handle);
  FILE *fp=fopen(buffer,"w");
  if(fp){
    fwrite(cb, cb->len, 1, fp);
    fclose(fp);
  }
#endif

#ifdef DEBUG_API
  fhputs("DEBUG_API: entering process_VRCB\n",vr->cfh);
#endif

  switch(cb->id)
    {

      /*
       * Adding Interface
       */

    case 1:
      {
#ifdef DEBUG_API
	fhputs("DEBUG_API: adding interface, Stage 1\n",vr->cfh);
#endif
	VRRB *rb=(VRRB*)malloc(1024);
	IF_INFORMATION_RB *irb=(IF_INFORMATION_RB*)rb;
	IF_ADDINTERFACE_CB *icb=(IF_ADDINTERFACE_CB*)cb;

	irb->handle=cb->handle;
	if(icb->len!=sizeof(IF_ADDINTERFACE_CB)){ 
	  irb->len=sizeof(IF_INFORMATION_RB);
	  irb->ifno=-1;
	  return log_rb(rb);
	}
#ifdef DEBUG_API
	fhputs("DEBUG_API: adding interface, Stage 2\n",vr->cfh);
#endif
	Interface *ifa=new Interface(vr, icb->ip, icb->nm, icb->bc, icb->name);

#ifdef DEBUG_API
	fhputs("DEBUG_API: adding interface, Stage 3",vr->cfh);
#endif

	irb->ifno=vr->add_interface(ifa);
	ifa->API_fill_info_rb(irb);
	return log_rb(rb);
      }


      /* 
       *	 Deleting Interfaces 
       */

    case 2:
      {
#ifdef DEBUG_API
	fhputs("DEBUG_API: delete interface: starting",vr->cfh);
#endif
	IF_DELETEINTERFACE_CB *icb=
	  (IF_DELETEINTERFACE_CB*)cb;
	IF_DELETEINTERFACE_RB *irb=
	  (IF_DELETEINTERFACE_RB*)malloc(sizeof(IF_DELETEINTERFACE_RB));
	irb->handle=icb->handle;
	irb->len=sizeof(IF_DELETEINTERFACE_RB);

	irb->ifno=0xffff;
	if(!vr->delete_interface(icb->ifno))
	  irb->ifno=icb->ifno;
	
#ifdef DEBUG_API
	fhputs("DEBUG_API: delete interface: finished",vr->cfh);
#endif

	return log_rb(irb);
      }

      /*
       *            query interface ids
       */

    case 3:
      {
	VRRB *rb=(VRRB*)malloc(1024);
	rb->handle=cb->handle;
#ifdef DEBUG_API
	fhputs("DEBUG_API: somebody queries the interface ids\n",vr->cfh);
#endif
	int i,j;
	rb->len=sizeof(VRRB)+sizeof(char)*vr->num_of_interfaces;
	char *body=(char*)rb+sizeof(VRRB);
	for(i=j=0;i<=vr->highest_interface_idx;++i){
	  if(vr->in[i]!=NULL)body[j++]=i;
	}
	return log_rb(rb);
      }


      /*
       * query interface by name
       */
      
    case 4:
      {
	IF_IDNAMEQUERY_CB *icb=(IF_IDNAMEQUERY_CB*)cb;
#ifdef DEBUG_API
	fhprintf(vr->cfh,"DEBUG_API: somebody queries the id for '%s'\n",icb->name);
#endif
	VRRB *rb=(VRRB*)malloc(1024);
	IF_INFORMATION_RB *irb=(IF_INFORMATION_RB*)rb;
	irb->handle=cb->handle;
	irb->len=sizeof(IF_INFORMATION_RB);
	irb->ifno=-1;
	  
	for(int i=0;i<=vr->highest_interface_idx;++i)
	  if(vr->in[i]!=NULL){
	    if(!strcmp(icb->name,vr->in[i]->id())){
	      vr->in[i]->API_fill_info_rb(irb);
	      irb->ifno=i;
	      break;
	    }
	  }
	return log_rb(rb);
      }

      
      /*
       * interface specific commands
       */
      
    case 10:	
      {
	IF_VRCB *icb=(IF_VRCB*)cb;
#ifdef DEBUG_API
	fhprintf(vr->cfh,"DEBUG_API: something interface specific with cmd %u\n",icb->cmd);
#endif
	if(icb->ifno>vr->highest_interface_idx || vr->in[icb->ifno]==NULL){ // error
#ifdef DEBUG_API
	  fhprintf(vr->cfh,"DEBUG_API: somewhat interface specific error\n");
#endif
	  VRRB *rb=(VRRB*)malloc(1024);
	  IF_INFORMATION_RB* irb=(IF_INFORMATION_RB*)rb;
	  irb->handle=icb->handle;
	  irb->len=sizeof(IF_INFORMATION_RB);
	  irb->ifno=-1;
#ifdef DEBUG_API
	  fhprintf(vr->cfh,"DEBUG_API: wrong interface id: %d\n",icb->ifno);
#endif
	  return log_rb(rb);
	}
	VRRB *rb=vr->in[icb->ifno]->APIconfigure(cb);
	return log_rb(rb);
      }
    case 20:			// routing 
      {
#ifdef DEBUG_API
	ROUTE_VRCB *icb=(ROUTE_VRCB*)cb;
	fhprintf(vr->cfh,"DEBUG_API: something routing specific with cmd %u\n",icb->command);
#endif
	VRRB *rb=vr->routing->APIconfigure(cb);
	return log_rb(rb);
      }
    case 21:			// list of scheduled events
      {
#ifdef DEBUG_API
	fhputs("DEBUG_API: query for event listing",vr->cfh);
#endif
	VRRB *rb=(VRRB*)malloc(sizeof(VRRB)+sizeof(struct EVENT_REC)*MAX_NUM_OF_EVENTS);
	struct EVENT_REC *ev = (struct EVENT_REC*)(rb+1);

	int j=0;
	for(int i=0;i<vr->vrsched->num_of_events;++i){
	  if(vr->vrsched->events[i]!=NULL){
	    ev[j].type=vr->vrsched->events[i]->type;
	    if(vr->vrsched->events[i]->suspended){
	      ev[j].suspend_time=vr->vrsched->events[i]->suspend_time;
	    } else {
	      ev[j].suspend_time=0.0;
	    }  
	    ev[j].fh_r=vr->vrsched->events[i]->fh_r;
	    ev[j].fh_w=vr->vrsched->events[i]->fh_w;
	    ev[j].hits=vr->vrsched->events[i]->hits;
	    strncpy(ev[j].comment,vr->vrsched->events[i]->o->id(),16);
	    ev[j].comment[15]=0;
	    ++j;
	  }
	}
	rb->len=sizeof(VRRB)+sizeof(struct EVENT_REC)*j;
	rb->handle=cb->handle;
	return log_rb(rb);
      }
    case 30:			// Filter commands
      {
#ifdef DEBUG_API
	fhputs("DEBUG_API: filter commands\n",vr->cfh);
#endif
	
	return log_rb(vr->pf->APIconfigure(this,cb));
      }
      
    case 40:			// loadable object
      {
#ifdef DEBUG_API
	fhputs("DEBUG_API: loadable object manipulation\n",vr->cfh);
#endif
	return log_rb(vr->vrsched->APIconfigure(this, cb));
      }
    case 50:			// IP packet encapsulated
      {
#ifdef DEBUG_API
	fhputs("DEBUG_API: got encapsulated IP packet\n",vr->cfh);
#endif
	return log_rb(ip_process(cb));
      }
    }
  return NULL;
}

unsigned short API::get_handle(){
  return handle++;
}


/*
 * API got an IP packet ...
 */

VRRB* API::ip_process(VRCB *v)
{
  Packet *p;
  struct ResultType {
    VRRB vrrb;
    unsigned short misc;
  };

  struct RequestType {
    VRCB vrcb;
    unsigned short flags;
    unsigned short mode;
    unsigned short misc;
  };
  
  RequestType *b=(RequestType*)v;
  p=(Packet*)(b+1);		// packet attached
  
  ResultType *rb=(ResultType*)malloc(sizeof(ResultType));
  rb->vrrb.handle=b->vrcb.handle;
  rb->vrrb.rb_type=0;		// sync
  rb->vrrb.len=sizeof(ResultType);
  

  if(b->flags&16){		// set fragmentation header
    p->frag=0;
  }

  if(b->flags&8){		// use default source address 
    int i;
    for(i=0;i<vr->num_of_interfaces;++i)
      if(vr->in[i]){
	p->source=vr->in[i]->get_ip();
	break;
      }
    if(i==vr->num_of_interfaces){
      rb->misc=0xffff;
      return (VRRB*)rb;
    }
  }

  if(b->flags&4){		// set ttl
    p->ttl=vr->ttl_default;
  }

  if(b->flags&2){		// recalculate crc
    p->crc=0;
    p->crc=cksum((char*)p, sizeof(PKT_HDRLEN(p)));
  }

  if(b->flags&1){		// check IP header
    if(ip_header_check(p,b->vrcb.len-sizeof(RequestType))){
      rb->misc=0xffff;
      return (VRRB*)rb;
    }
  }

 
  switch(b->mode)
    {
    case 0:	
      /* nur routen 
       */
      rb->misc=vr->routing->route(p)&0xff;
      return (VRRB*)rb;
    case 1:
      /* komplettes redistribute
       * da intern, interface nummer ist NULL
       */
      rb->misc=vr->pf->distribute(NULL, p);
      return (VRRB*)rb;
    case 2:		
      /* interface forced
       */
      if(vr->in[b->misc]!=NULL){
	vr->in[b->misc]->send_packet(p);
	rb->misc=b->misc;
	return (VRRB*)rb;
      }
      rb->misc=0xffff;
      return (VRRB*)rb;
    }
  return NULL;
}


VRRB* API::log_rb(VRRB *rb)
{
#ifdef LOG_APICOM
  char buffer[256];
  sprintf(buffer,"%s%d.rb",LOG_APICOM,rb->handle);
  FILE *fp=fopen(buffer,"w");
  if(fp){
    fwrite(rb, rb->len, 1, fp);
    fclose(fp);
  }
#endif

  rb->rb_type=0;			// synchronous traffic
  return rb;
}

int API::send_RB(VRRB *r)
{
  return write(api_msg[1],r, r->len);
}


/*
 * if a filter is setup and no callback
 * function specified, this funtion is called
 */

int API::forward(void *v, Packet *p)
{
  FFilter *f = (FFilter*)v;
  
  VRRB rb;
  rb.rb_type=1;			// filter record
  rb.handle=f->cb_handle;
  rb.len=sizeof(VRRB)+PKT_LENGTH(p);
  write(api_msg[1],&rb,sizeof(VRRB));
  write(api_msg[1],p,PKT_LENGTH(p));
  return 0;
}

/*
 * if the API implements an Protocolstack
 * this funtion is called
 */

int API::process(PSPacket *psp)
{
  VRRB rb;
  rb.rb_type=2;			// Protocolstack packet
  rb.handle=0;
  rb.len=sizeof(VRRB)+psp->pktlen;
  write(api_msg[1],&rb,sizeof(VRRB));
  write(api_msg[1],&(psp->pkt),psp->pktlen);
  return 0;
}



