#ifndef _api_h_
#define _api_h_

class VRevent;

#include "vr.H"
#include "vrsched.H"

//> VRCB
struct VRCB 
{
  unsigned short handle;	//- to this request will be sent back containing this handle.
  unsigned short len;		//- the len of the complete Control Block
  unsigned short id;		//- A Command id, spcifying the action to be taken 
};
//<

//> VRRB
struct VRRB 
{
  unsigned short handle;	//- the handle of the according VRCB
  unsigned short rb_type;	//- the type of the returned VRRB 
				//- 0 for synchronous traffic 
				//- 1 for filter record with handle for filter handle 
				//- 2 for a protocol stack packet 
  unsigned short len;		//- the len of the complete VRRB
};
//<

//> EVENT_REC
struct EVENT_REC
{
  unsigned short type;		//- which event (IO_READ, IO_WRITE) 
  unsigned short thrown;	//- what happend IO_READ, IO_WRITE
  double suspend_time;		//- event suspended for this time
				//- is 0 if not suspended 
  unsigned short fh_r;		//- the read handle assoc. with the event 
  unsigned short fh_w;		//- the write handle assoc with the event 
  char comment[16];		//- name of the object the event belongs to
  unsigned long hits;		//- how many hits had the event  
  int forced;			//- force interupt call 
};
//<

//> LOB_INFO_BLOCK
struct LOB_INFO_BLOCK
{
  unsigned short lob_id;	//- the if of the loadable object
  unsigned short mode;		//- the mode the lob is in 
  unsigned long size;		//- the size of the object code
  double uptime;		//- the time the load is loaded
  char name[32];		//- the name of the lob 
};
//<


#include "api_interface.H"
#include "api_routing.H"
#include "api_filter.H"
#include "forwarder_if.H"

//> API
class API : public virtual VRschedObj, public virtual ForwarderIF, public virtual ProtocolStack
{
  friend class LOBenv;

public:
  API(VR*,int mode=1);		//- the constructor mode=0 for passive mode  
  API(VR*,int fh_r, int fh_w);	//- the constructor mode=0 for passive mode  

  ~API();			//- the desctructor 
  char *id(){return idstring;}	//- an id string
  int interrupt(VRevent *e);	//- is called by the Event scheduler 
  unsigned short get_handle();	//- get a unique handle id

  int write_CB(VRCB *);		//- write a CB to the API
  int read_CB(VRRB *,int);	//- read a CB from the API
  int send_RB(VRRB *);		//- send an asynchronous RB to the API 

  int forward(void *, Packet *);
  int process(PSPacket*);

private:

  VRCB *cb;			//- buffer to keep the VRCB
  int api_pipe[2];
  int api_msg[2];
  unsigned short handle;
  int api_modus;		//- is true if API handles own pipes
protected:

  VRevent *event;		//- the event for the scheduler
  char *idstring;
  VR* vr;			//- the Virtuel Router 
  VRRB *ip_process(VRCB *cb); //- process encapsulated ip packets
  VRRB *process_VRCB(VRCB *cb);
  VRRB *log_rb(VRRB*);		//- wrapper to allow the loging
				//- of RBs
};
//<


#endif


