#include <string.h>
#include <stdlib.h>
#include "vr.H"
#include "forwarder.H"
#include "tools.H"
#include "api.H"
#include "packet.H"
#include "ffilter.H"
#include "interface.H"
#include "ioc.H"
#include "icmp.H"
#include "tunnel_endpoint.H"
#include "udp.H"
#include "routing.H"
#include "defragmenter.H"
#include "vrevent.H"
#include "socks.H"


int Forwarder::interrupt(VRevent *e)
{
  char buffer[2048];
  Packet *p=(Packet*)buffer;
  
  if(e->thrown&IO_READ){
    read(hostif[0], buffer, 20);
    read(hostif[0], buffer+20,PKT_LENGTH(p)-20);
#ifdef DEBUG_FORWARDER
    fhprintf(vr->cfh, "DEBUG_FWD host layer: %s\n",pkt2char(p));
#endif
    distribute(NULL,p);
  }
  return 0;
}

    
int Forwarder::sort()
{
  for(int d=1; d;){
    d=0;
    for(int i=0;i<last-1;++i){
      if(table[i] && table[i+1] && table[i]->prio < table[i+1]->prio){
	FFilter *t;t=table[i]; table[i]=table[i+1]; table[i+1]=t;
	++d;
      }
    }
  }
  return 0;
}



int Forwarder::add(FFilter *f)
{
  for(int i=0;i<size_of_table;++i){
    if(!table[i]){
      table[i]=f;
      if(i>=last)last=i+1;
      num_of_filters++;
      sort();
      return 0;
    }

  }
  return 1;
}



int Forwarder::del(FFilter *f)
{
  for(int i=0;i<last;++i){
    if(table[i] && table[i]->equals(f)){
      delete table[i];
      table[i]=NULL;
      num_of_filters--;
      if(i>=last-1)last--;
      sort();
      return 0;
    }
  }
  return 1;
}



Forwarder::Forwarder(VR *v, int n)
{
  vr=v;
  last=0;
  num_of_filters=0;
  size_of_table=n;
  table=(FFilter**)malloc(sizeof(FFilter*)*size_of_table);
  memset(table,0,size_of_table*sizeof(FFilter*));
  memset(host_protocols,0,256*sizeof(ProtocolStack*));
  host_protocols[1]=vr->icmpstack;
  host_protocols[4]=vr->tunnel_ep;
  host_protocols[17]=vr->udpstack;
  dfrg=new Defragmenter();
  pipe(hostif);
  raw_socks=new RawSock(vr);
  udp_socks=new UDPSock(vr);
  
  event = new VRevent(this);
  event->type=IO_READ;
  event->suspended=0;
  event->thrown=0;
  event->suspend_time=HUGE_VAL;
  event->fh_r=hostif[0];
  event->fh_w=hostif[1];
  vr->vrsched->notify(event);

  return;
}



Forwarder::~Forwarder()
{
  delete dfrg;
  delete table;
  return;
}



FFilter *Forwarder::lookup(Packet *p)
{
  for(int i=0;i<last;++i){
    if(table[i] 
       && !(table[i]->flags&FFILTER_FLAG_DELETE) 
       && table[i]->matches(p))return table[i];
  }
  return NULL;
}

/* the packet and the interface it is coming from
  if the distribute function is called to send a
  locally generated packet, the interface is NULL
 
  return codes:
 
  low byte        n
  high byte       code
 
  codes:
           0 packet routed to interface 
             n != 0xff  interface number
             n =  0xff  error (no route) 
           1 local protocolstack
               n =  0 okay
               n != 0 error
           2 packet dropped due to ttl 
               n=ttl
	   0xff
	       n=0xff error
 */

unsigned short Forwarder::distribute(Interface *ifc, Packet *p)
{
  int drop_it=0;
  

#ifdef DEBUG_FORWARDER
  fhprintf(vr->cfh, "DEBUG_FWD in %s\n",pkt2char(p));
#endif


  // check for filters with an filter priority < 1000
  //

  drop_it=0;
  for(int i=0;i<last;++i){
    if(table[i] && table[i]->prio<1000 && table[i]->matches(p) && table[i]->forwarder){
      if(!((table[i]->mode)&2)){	// no defragment
	table[i]->forwarder->forward(table[i],p);
      }
      else {
	PSPacket *pc=dfrg->defragment(p);
	if(pc){			// last fragment
	  table[i]->forwarder->forward(table[i],&pc->pkt);
  	}
      }
      if((table[i]->mode)&1){	// drop it
	drop_it++;
	break;  
      }
    }
  }

  if(drop_it){	
#ifdef DEBUG_FORWARDER
    fhprintf(vr->cfh, "DEBUG_FWD packet %s was removed due to filter I\n",pkt2char(p));
#endif
    return 0x2001;		// packet dropped, due to filter
  }

  // what belongs to this machine is sent to
  // the appropriate protocol handlers
  // The protocol stacks

  for(int i=0;i<vr->num_of_interfaces;++i)
    if(vr->in[i] && vr->in[i]->get_ip()==PKT_DEST(p)){
#ifdef DEBUG_FORWARDER_X
      fhputs("DEVUG_FWD packet for local host\n",vr->cfh);
#endif

      PSPacket *pc=dfrg->defragment(p);
      
      if(pc){
	if(host_protocols[PKT_PROTO(p)]){
	  host_protocols[pc->proto]->process(pc);
	  return 0x1000;
	} 
	else{
#ifdef DEBUG_FORWARDER_X
	  fhputs("DEBUG_FWD no protocol stack -> raw_socket_dispatch\n",vr->cfh);
#endif
	  raw_socks->dispatch(p);
	  return 0x1000 + 0x1;
	}
      }
      return 0xffff;
    }


  // check for filters with an filter priority >= 1000
  //
  
  drop_it=0;
  for(int i=0;i<last;++i){
    if(table[i] && table[i]->prio>=1000 && table[i]->matches(p) && table[i]->forwarder){
      if(!((table[i]->mode)&2)){	// no defragment
	table[i]->forwarder->forward(table[i],p);
      }
      else {
	PSPacket *pc=dfrg->defragment(p);
	if(pc){			// last fragment
	  table[i]->forwarder->forward(table[i],&pc->pkt);
  	}
      }
      if((table[i]->mode)&1){	// drop it
	drop_it++;
	break;  
      }
    }
  }

  if(drop_it){	
#ifdef DEBUG_FORWARDER
    fhprintf(vr->cfh, "DEBUG_FWD packet %s was removed due to filter I\n",pkt2char(p));
#endif
    return 0x2001;		// packet dropped, due to filter
  }

  p->ttl=p->ttl-1;
  
  if(!p->ttl){			// ICMP senden
#ifdef DEBUG_FORWARDER_X
    fhprintf(vr->cfh, "DEBUG_FWD drop due to ttl exceed\n");
    fhprintf(vr->cfh, "DEBUG_FWD %s\n",pkt2char(p));
#endif
    if(ifc!=NULL)vr->icmpstack->ttl_exceeded(ifc,p);
    return 0x2000;
  }

  // at the end: routing 

#ifdef DEBUG_FORWARDER
  fhputs("DEBUG_FWD routing packet\n",vr->cfh);
#endif

  return (vr->routing->route(p))&0xfff;
}




//
// The API call
//


VRRB* Forwarder::APIconfigure(API *source_api, VRCB *in)
{
  /*
   * Protocol Stacks
   */

  if(in->id==31){		// configuration of Protocol Stacks
    struct PROTOSTACK_CB *cb=(struct PROTOSTACK_CB*)in;
    PROTOSTACK_RB *rb=(PROTOSTACK_RB*)malloc(sizeof(PROTOSTACK_RB));
    rb->handle=cb->handle;
    rb->rb_type=0;		// synchronous
    rb->len=sizeof(PROTOSTACK_RB);
    
    if(host_protocols[cb->protocol]!=NULL){
      rb->protocol=0;
      return (VRRB*)rb;
    }

    if(cb->stack==NULL)cb->stack=source_api;

    host_protocols[cb->protocol]=cb->stack;
   
    return (VRRB*)rb;
  }
  
  /*
   * Filter setup and control
   &
  */


  if(in->id==30){		// Filter configuration
    struct FILTER_CB *cb=(struct FILTER_CB*)in;
  
    switch(cb->filter_command)
      {
      case 0:			// get list of filters
	{
	  struct dummy0 : public VRRB{
	    struct FILTER_REC filter;
	  };
	  
	  struct dummy0 *rb=(struct dummy0*)
	    malloc(sizeof(VRRB)+sizeof(struct FILTER_REC)*num_of_filters);
	  rb->len=sizeof(VRRB)+sizeof(struct FILTER_REC)*num_of_filters;
	  rb->handle=cb->handle;
	  
	  FILTER_REC *f=&(rb->filter);
	  for(int i=0;i<num_of_filters;++i){
	    f[i].prio=table[i]->prio;
	    f[i].forwarder=table[i]->forwarder;
	    f[i].source_ip=table[i]->source_ip;
	    f[i].source_nm=table[i]->source_nm;
	    f[i].dest_ip=table[i]->dest_ip;
	    f[i].dest_nm=table[i]->dest_nm;
	    f[i].tos=table[i]->tos;
	    f[i].proto=table[i]->proto;
	    f[i].ipopt_type=table[i]->ipopt_type;
	    f[i].ipopt_value=table[i]->ipopt_value;
	    f[i].mode=table[i]->mode;
	    f[i].cb_handle=table[i]->cb_handle;
	    strncpy(f[i].name,table[i]->name,31); f[i].name[31]=0;
	  }
	  return rb;
	}
	
      case 2:			// add FILTER
	{		
	  struct FILTER_ADDDEL_CB *cb=(struct FILTER_ADDDEL_CB *)in;
#ifdef DEBUG_API
	  fhprintf(vr->cfh,"DEBUG_API: API adds filter '%s'\n",cb->filter.name);
#endif
	  
	  // create filter instance and copy specs
	  
	  FFilter *f = new FFilter();
	  f->prio=cb->filter.prio;
	  f->forwarder=cb->filter.forwarder;
	  if(f->forwarder==NULL){
	    f->forwarder=source_api;
	  }
	  f->source_ip=cb->filter.source_ip;
	  f->source_nm=cb->filter.source_nm;
	  f->dest_ip=cb->filter.dest_ip;
	  f->dest_nm=cb->filter.dest_nm;
	  f->tos=cb->filter.tos;
	  f->proto=cb->filter.proto;
	  f->ipopt_type=cb->filter.ipopt_type;
	  f->ipopt_value=cb->filter.ipopt_value;
	  f->mode=cb->filter.mode;
	  strncpy(f->name,cb->filter.name,31); f->name[31]=0;
	  f->cb_handle=cb->handle;
	  
	  // prepare result 
	  
	  struct dummy0 : public VRRB{
	    struct FILTER_REC filter;
	  };
	  
	  struct dummy0 *rb=(struct dummy0*)malloc(sizeof(struct dummy0));
	  rb->len=sizeof(struct dummy0);
	  rb->handle=cb->handle;
	  memcpy(&rb->filter, &cb->filter,sizeof(FILTER_REC));
	  
	  rb->filter.cb_handle=f->cb_handle;
	  
	  // add filter
	  
	  if(!add(f)){
#ifdef DEBUG_API
	    fhprintf(vr->cfh,"DEBUG_API: filter '%s' add okay\n",cb->filter.name);
#endif
	    return rb;		// okay
	  }
	  
#ifdef DEBUG_API
	  fhprintf(vr->cfh,"DEBUG_API: filter add failed\n");
#endif
	strcpy(rb->filter.name,"add failed");
	rb->filter.forwarder=NULL;
	rb->filter.mode=0;	// some errorcode
	return rb;
	}
      }
  }
  
  return NULL;
}
