#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <intercon/udp_connection.H>

#include "api.H"
#include "vr.H"
#include "qcroot.H"
#include "forwarder.H"
#include "tools.H"
#include "packet.H"
#include "routing.H"
#include "vrevent.H"
#include "interface.H"

Interface::Interface(VR *n, unsigned long ipp, unsigned long nmp, unsigned long bcp, char *name)
{
  vr=n;
  ip=ipp; nm=nmp; bc=bcp;
  idstring=strdup(name);
  constructor_common();
  return;
}
Interface::Interface(VR *n, char *ipaddr, char *netmask, char *name)
{
  struct in_addr inp;
  vr=n;
  inet_aton(ipaddr, &inp);ip=inp.s_addr;
  inet_aton(netmask,&inp);nm=inp.s_addr;
  idstring=strdup(name);
  constructor_common();
  return;
}
 
void Interface::constructor_common()
{
  *last_in_buffer=0;
  *last_out_buffer=0;

  iptrans_enable=vr->iptrans_enable;

  iptrans_rx_ip=vr->iptrans_rx_address;
  iptrans_rx_nm=vr->iptrans_rx_netmask;
  iptrans_rx_value=vr->iptrans_rx_value;
  iptrans_rx_pattern=vr->iptrans_rx_mask;

  iptrans_tx_ip=vr->iptrans_tx_address;
  iptrans_tx_nm=vr->iptrans_rx_netmask;
  iptrans_tx_value=vr->iptrans_tx_value;
  iptrans_tx_pattern=vr->iptrans_tx_mask;

  //  char2ipnm("172.0.0.0/8",&iptrans_rx_ip, &iptrans_rx_nm);
  //  char2ipnm("10.0.0.0/8",&iptrans_tx_ip, &iptrans_tx_nm);
  //  char2ipnm("10.0.0.0/8",&iptrans_rx_value, &iptrans_rx_pattern);
  //  char2ipnm("172.0.0.0/8",&iptrans_tx_value, &iptrans_tx_pattern);

  forwarder=NULL;
  sender=NULL;
  receiver=NULL;
  fh_r=fh_w=-1;
  q_send=new QCroot(vr);
  connectiontype=disconnected;
  cable_no=0;
  cable_end=0;

  event = new VRevent(this);
  packet_in = (Packet*)malloc(2048);
  rx_stat=0;
  tx_stat=0;
  err_stat=0;

  bandwidth=vr->if_bandwidth;
  bucketsize=vr->if_bucketsize;

  last_call=gettimeasdouble();
  bucket_write=0;
  bucket_read=0;
  return;
}

Interface::~Interface()
{
  DisConnect();
  delete event;
  delete q_send;
  free(packet_in);
  free(idstring);
  return;
}

/*
 * This function handles all interface configuration
 * it may only be called, when the base system is paused 
 * (e.g. from am event slot)
 *
 */
VRRB* Interface::APIconfigure(VRCB *v)
{
  IF_VRCB* in=(IF_VRCB*)v;

  if(in==NULL){
    return NULL;
  }
  char *body=(char*)in+sizeof(IF_VRCB); // zeigt auf die Payload des CB
  
  switch(in->cmd)
    {
    case 1:     // Information Query
      {
	if(in->len!=sizeof(IF_VRCB))break;
	IF_INFORMATION_RB* r=(IF_INFORMATION_RB*)malloc(sizeof(IF_INFORMATION_RB));
	API_fill_info_rb(r);
	r->ifno=in->ifno;
	r->handle=in->handle;
	return (VRRB*)r;
      }
    case 2:     // Reset Statistic counters
      {
	if(in->len!=sizeof(IF_VRCB))break;
	rx_stat=0;
	tx_stat=0;
	err_stat=0;
	drop_stat=0;
	IF_INFORMATION_RB* r=(IF_INFORMATION_RB*)malloc(sizeof(IF_INFORMATION_RB));
	API_fill_info_rb(r);
	r->ifno=in->ifno;
	r->handle=in->handle;
	return (VRRB*)r;
      }
    case 101: // change interface name
      {
	if(in->len!=sizeof(IF_VRCB)+10)break;
	free(idstring);
	idstring=strdup(body);
	IF_INFORMATION_RB* r=(IF_INFORMATION_RB*)malloc(sizeof(IF_INFORMATION_RB));
	API_fill_info_rb(r);
	r->ifno=in->ifno;
	r->handle=in->handle;
	return (VRRB*)r;
      }
    case 102: // change ip, nm and bc;
      {
	struct dummy0 : public IF_VRCB{
	  unsigned long ip;
	  unsigned long nm;
	  unsigned long bc;
	};
	struct dummy0 *dummy_cb=(struct dummy0*)in;
	if(dummy_cb->len!=sizeof(struct dummy0))break;

	if(dummy_cb->ip!=0xffffffff)ip=dummy_cb->ip;
	if(dummy_cb->nm!=0xffffffff)nm=dummy_cb->nm;
	if(dummy_cb->bc!=0xffffffff)bc=dummy_cb->bc;

	IF_INFORMATION_RB* r=(IF_INFORMATION_RB*)malloc(sizeof(IF_INFORMATION_RB));
	API_fill_info_rb(r);
	r->ifno=in->ifno;
	r->handle=in->handle;
	return (VRRB*)r;
      }
    case 105: // change bandwidth and bucketsize
      {
	struct dummy0 : public IF_VRCB{
	  unsigned long bw;
	  unsigned long bs;
	};
	  
	struct dummy0 *dummycb=(struct dummy0*)in;
	
	if(dummycb->len!=sizeof(struct dummy0))break;
	//	printf("bandwidth was: %lu %lu\n", bandwidth, bucketsize);

	bandwidth=dummycb->bw; 
	if(dummycb->bs!=0xffffffff)bucketsize=dummycb->bs;
	bucket_write=bucketsize;
	

	//	printf("bandwidth is: %lu %lu\n", bandwidth, bucketsize);

	IF_INFORMATION_RB* r=(IF_INFORMATION_RB*)malloc(sizeof(IF_INFORMATION_RB));
	API_fill_info_rb(r);
	r->ifno=in->ifno;
	r->handle=in->handle;
	return (VRRB*)r;
      }
    case 107: // change iptrans rx 
      {
	if(in->len!=sizeof(IF_VRCB)+4*sizeof(unsigned long))break;

	iptrans_rx_ip=((unsigned long*)body)[0];
	iptrans_rx_nm=((unsigned long*)body)[1];
	iptrans_rx_value=((unsigned long*)body)[2];
	iptrans_rx_pattern=((unsigned long*)body)[3];

	IF_INFORMATION_RB* r=(IF_INFORMATION_RB*)malloc(sizeof(IF_INFORMATION_RB));
	API_fill_info_rb(r);
	r->ifno=in->ifno;
	r->handle=in->handle;
	return (VRRB*)r;
      }
    case 108: // change iptrans rx 
      {
	if(in->len!=sizeof(IF_VRCB)+4*sizeof(unsigned long))break;

	iptrans_tx_ip=((unsigned long*)body)[0];
	iptrans_tx_nm=((unsigned long*)body)[1];
	iptrans_tx_value=((unsigned long*)body)[2];
	iptrans_tx_pattern=((unsigned long*)body)[3];

	IF_INFORMATION_RB* r=(IF_INFORMATION_RB*)malloc(sizeof(IF_INFORMATION_RB));
	API_fill_info_rb(r);
	r->ifno=in->ifno;
	r->handle=in->handle;
	return (VRRB*)r;
      }
    case 109: // connect interface
      {
	struct dummy : public IF_VRCB
	{
	  char ctype;			// type of connection
	  unsigned long cval[3];
	};
	struct dummy *cb=(struct dummy*)in;

	if(in->len!=sizeof(struct dummy))break;

	switch(cb->ctype)
	  {
	  case 0: // disconnect
	    DisConnect();
	    break;
	  case 1: // softlink device
	    ConnectToSoftlink(cb->cval[1]);
	    break;
	  case 2: // local
	    ConnectToFIFO(cb->cval[1], cb->cval[2]);
	    break;
	  case 3: // UDP
	    ConnectToUDP(cb->cval[0], cb->cval[1], cb->cval[2]);
	    break;
	  case 4: // IPIP tunnel
	    Setup_IPIPtunnel(cb->cval[0], cb->cval[2]);
	    break;
	  }
	IF_INFORMATION_RB* r=(IF_INFORMATION_RB*)malloc(sizeof(IF_INFORMATION_RB));
	API_fill_info_rb(r);
	r->ifno=in->ifno;
	r->handle=in->handle;
	return (VRRB*)r;
      }
    case 120: // Queueing System API is handled by the queueing system
#ifdef DEBUG_API
	fhprintf(vr->cfh,"DEBUG_API: accessing queueing System API\n");
#endif
      VRRB* r=q_send->APIconfigure(v);
      r->handle=in->handle;
      return r;
      break;
    }

  /*
   * return ERROR (ifno == -1)
   * 
   */

  IF_INFORMATION_RB* r=(IF_INFORMATION_RB*)malloc(sizeof(IF_INFORMATION_RB));
  API_fill_info_rb(r);
  r->ifno=-1; 
  r->handle=in->handle;
  return (VRRB*)r;
}

int Interface::API_fill_info_rb(IF_INFORMATION_RB *r)
{  
  if(r==NULL)return 0;
  memset(r,0,sizeof(IF_INFORMATION_RB));
  r->len=sizeof(IF_INFORMATION_RB);
  r->ip=ip;
  strcpy(r->name,idstring);
  r->netmask=nm;
  r->broadcast=bc;
  r->rx=rx_stat;
  r->tx=tx_stat;
  r->errors=err_stat;
  r->drops=drop_stat;
  r->bandwidth=bandwidth;
  r->bucketsize=bucketsize;
  r->bucketlevel=bucket_write;

  if(iptrans_enable){
    r->iptrans_rx_ip      =iptrans_rx_ip;
    r->iptrans_rx_nm      =iptrans_rx_nm;
    r->iptrans_rx_value   =iptrans_rx_value;
    r->iptrans_rx_pattern =iptrans_rx_pattern;
    r->iptrans_tx_ip      =iptrans_tx_ip;
    r->iptrans_tx_nm      =iptrans_tx_nm;
    r->iptrans_tx_value   =iptrans_tx_value;
    r->iptrans_tx_pattern =iptrans_tx_pattern;
  }

  switch(connectiontype){
  case disconnected:
    r->ctype=0;
    r->cval[0]=r->cval[1]=r->cval[2]=0;
    break;
  case softlink:
    r->ctype=1;
    r->cval[0]=r->cval[2]=0;
    r->cval[1]=cable_no;
    break;
  case local:
    r->ctype=2;
    r->cval[0]=0;
    r->cval[1]=cable_no;
    r->cval[2]=cable_end;
    break;
  case tunnel:
    r->ctype=3;
    r->cval[0]=sender->get_dest_addr();
    r->cval[1]=sender->get_dest_port();
    r->cval[2]=receiver->get_port();
    break;
  case ipiptunnel:
    r->ctype=4;
    r->cval[0]=ipip_destination;
    r->cval[1]=0;
    r->cval[2]=ipip_mtu;
    break;
  }
  return 0;
}

/* new connect to a Softlink device
 *
 * return values:
 * 0 device is now disconnected
 * 1 was not connected
 * 2 unkown connectio type (strange);
 *
 */

int Interface::DisConnect()
{
  switch(connectiontype)
    {
    case disconnected:
      return 1;
    case softlink:
      vr->vrsched->release(event->event_id);
      event->clear();
      close(fh_r); 
      close(fh_w); 
      fh_r=fh_w=-1;
      connectiontype=disconnected;
      cable_no=0;
      cable_end=0;
      return 0;

    case local:
      vr->vrsched->release(event->event_id);
      event->clear();
      close(fh_r); 
      close(fh_w); 
      fh_r=fh_w=-1;
      connectiontype=disconnected;
      cable_no=0;
      cable_end=0;
      return 0;

    case tunnel:
      vr->vrsched->release(event->event_id);
      event->clear();
      delete sender;
      delete receiver;
      sender=NULL;
      receiver=NULL;
      fh_r=fh_w=-1;
      connectiontype=disconnected;
      cable_no=0;
      cable_end=0;
      return 0;

    case ipiptunnel:
      connectiontype=disconnected;
      ipip_mtu=0;
      ipip_destination=0;
      return 0; 
    }
  return 2;
}

/* new connect to a Softlink device
 *
 * return values:
 * 0 okay
 * 1 already connected
 * 2 cannot open device
 *
 */
int Interface::ConnectToSoftlink(unsigned short id)
{
  if(connectiontype!=disconnected){
    return 1;
  }

  char buff[256];
  sprintf(buff,"%s%d",vr->softlink_basename,id);
  fh_w=fh_r=open(buff,O_RDWR);
  if(fh_r>0) {
    event->type=IO_READ|IO_WRITE;
    event->suspended=0;
    event->suspend_time=HUGE_VAL;
    event->thrown=0;
    event->fh_r=fh_r;
    event->fh_w=fh_w;
    vr->vrsched->notify(event);
    connectiontype=softlink;
    cable_no=id;
    return 0;
  }

  return 2;
}


/* new connect to a FIFO device
 *
 * return values:
 * 0 okay
 * 1 already connected
 * 2 wrong cable ends
 * 3 cannot connect
 */
int Interface::ConnectToFIFO(unsigned short cid, unsigned short cend)
{
  char e;
  if(connectiontype!=disconnected)return 1;
  if(cend!=1 && cend!=2)return 2;
  if(cend==1)e='r'; else e='l';

  cable_no=cid;

  if(e=='r'){
    char buffr[32], buffl[32];
    sprintf(buffr,"%s%dr",vr->fifo_basename,cid);
    fh_w=open(buffr,O_RDWR|O_NONBLOCK);
    sprintf(buffl,"%s%dl",vr->fifo_basename,cid);
    fh_r=open(buffl,O_RDWR|O_NONBLOCK);
    cable_end=1;
  }
  if(e=='l'){
    char buffr[32], buffl[32];
    sprintf(buffr,"%s%dr",vr->fifo_basename,cid);
    fh_r=open(buffr,O_RDWR|O_NONBLOCK);
    sprintf(buffl,"%s%dl",vr->fifo_basename,cid);
    fh_w=open(buffl,O_RDWR|O_NONBLOCK);
    cable_end=2;
  }
  

  if(fh_w!=-1 && fh_r!=-1){
    event->fh_r=fh_r;
    event->fh_w=fh_w;
    event->type=IO_READ|IO_WRITE;
    event->suspended=0;
    event->suspend_time=HUGE_VAL;
    event->thrown=0;
    vr->vrsched->notify(event);
    connectiontype=local;
    return 0;
  }
  close(fh_r); close(fh_w);
  return 3;
}


/* new connect via UDP
 *
 * return values:
 * 0 okay
 * 1 already connected
 */
int Interface::ConnectToUDP(unsigned long to, unsigned short tx_port, unsigned short rx_port)
{
  if(connectiontype!=disconnected)return 1;

  sender = new UDP_Sender(to, tx_port);
  receiver = new UDP_Receiver(rx_port);

  sender->set_output_interval(0);
  receiver->set_output_interval(0);

  event->type=IO_READ|IO_WRITE;
  event->suspended=0;
  event->thrown=0;
  event->suspend_time=HUGE_VAL;
  event->fh_r=receiver->get_fh();
  event->fh_w=sender->get_fh();
  vr->vrsched->notify(event);
  connectiontype=tunnel;

  return 0;
}


int Interface::interrupt(VRevent *e)
{

  if(e->thrown&IO_READ){
    
    int len;
    if(!receiver){
      len=read(fh_r, packet_in, 2048);
    }
    else {
      len=receiver->receivebuff((char*)packet_in, 2048);
    }
    ++rx_stat;
    
#ifdef DEBUG_PACKETTRANSMISSION
    fhprintf(vr->cfh, "--- DEBUG_PACKETTRANSMISSION ---\n");
    fhprintf(vr->cfh, "incoming on interface %s\n",idstring);
    fhprintf(vr->cfh, "%s\n",pkt2char(packet_in));
#endif
    
#ifdef ENABLE_DUPLICATE_DETECTION
    if(*last_in_buffer && !bcmp(last_in_buffer,packet_in,20)){
      fhprintf(vr->cfh,"--- ERROR --- \n");
      fhprintf(vr->cfh,"duplicate packet received from link on interface %s\n",id());
      fhprintf(vr->cfh,"%s\n",pkt2char(packet_in));
    }
    memcpy(last_in_buffer,packet_in,20);
#endif ENABLE_DUPLICATE_DETECTION
    
    if(len==0){
      fhprintf(vr->cfh,"ERROR: read event, but no data\n");
      goto write_pkt;
    }
    
    //
    // sanity checks for softlink device

    if(connectiontype==softlink){
      unsigned int pl=(unsigned short)PKT_LENGTH(packet_in);
      if((unsigned)len<pl){
	++err_stat; 
	fhprintf(vr->cfh,"ERROR: too few data for packet header (%d bytes)\n",len);
	fhprintf(vr->cfh,"ERROR:  %s\n", pkt2char(packet_in));
	goto write_pkt;
      }
      if(cksum((char*)packet_in, PKT_HDRLEN(packet_in))){
	fhprintf(vr->cfh,"ERROR: len mismatch and checksum error (%d bytes)\n",len);
	fhprintf(vr->cfh,"ERROR: %s\n", pkt2char(packet_in));
      }
    }

    // new addresstranslation
      
    if(iptrans_enable && connectiontype==softlink && 
       ((PKT_DEST(packet_in)&iptrans_rx_nm) == (iptrans_rx_ip&iptrans_rx_nm))){
      unsigned long inv_nm=iptrans_rx_pattern^0xffffffff;
      unsigned long na=(PKT_DEST(packet_in)&inv_nm) | 
	(iptrans_rx_value&iptrans_rx_pattern);
#ifdef DEBUG_PACKETTRANSMISSION
      fhprintf(vr->cfh,"DEBUG_INF %s\n",idstring);
      fhprintf(vr->cfh,"DEBUG_INF in/dest address translation %s",n2char(PKT_DEST(packet_in)));
      fhprintf(vr->cfh,"->%s\n",n2char(na));
#endif
      PKT_DEST(packet_in)=na;
    }
    
    vr->pf->distribute(this, packet_in);
    
  }

write_pkt:

  if(e->thrown&IO_WRITE){
    double ti=gettimeasdouble();
    Packet *p;
    
    bucket_write+=(unsigned long)((ti-last_call)*bandwidth);
    last_call=ti;
    if(bucket_write>bucketsize)bucket_write=bucketsize;
    
    p=q_send->root_nextdeque();

    //
    // if no packet in queues, suspend write event 
    // the event is activated, when a packet is put to the queueing system
    //
    // if we got no packet, but packets remain in queue, suspend for 5 ms
    //

    if(p==NULL){
      e->suspended|=IO_WRITE;
      if(q_send->get_packets_in_qs())
	e->suspend_time=ti+0.001;
      else
	e->suspend_time=HUGE_VAL;
      
#ifdef DEBUG_EVENTSYSTEM  
      fhprintf(vr->cfh,"DEBUG_INF %s no packet write event suspended\n",idstring);
#endif
      return 0;
    }
    
    //
    // suspend for a short time, when the interfaces 
    // bandwidth limiter is empty 
    //

    if(bucket_write<(unsigned long)PKT_LENGTH(p)) {
      e->suspended|=IO_WRITE;
      e->suspend_time=ti+0.001;

#ifdef DEBUG_EVENTSYSTEM  
      fhprintf(vr->cfh,"DEBUG_INF %s bucket empty write event suspended\n",idstring);
#endif
      return 0;
    }    

    p=q_send->root_deque();

    bucket_write-=PKT_LENGTH(p);

    //
    // sanity checks for queuing system
    //


#ifdef ENABLE_DUPLICATE_DETECTION
    if(*last_out_buffer && !bcmp(last_out_buffer,p,20)){
      fhprintf(vr->cfh,"--- ERROR --- \n");
      fhprintf(vr->cfh,"duplicate packet from Queueing System on interface %s \n",id());
      fhprintf(vr->cfh,"%s\n",pkt2char(p));
    }
    memcpy(last_out_buffer,p,20);
#endif ENABLE_DUPLICATE_DETECTION

    //
    // new addresstranslation
    //

    if(iptrans_enable && connectiontype==softlink &&
       ((PKT_SOURCE(p)&iptrans_tx_nm) == (iptrans_tx_ip&iptrans_tx_nm))){
      unsigned long inv_nm=iptrans_tx_pattern^0xffffffff;
      unsigned long na=(PKT_SOURCE(p)&inv_nm) | (iptrans_tx_value&iptrans_tx_pattern);

#ifdef DEBUG_ADDRESSTRANSLATION
   fhprintf(vr->cfh, "--- DEBUG_ADDRESSTRANSLATION ---\n");
   fhprintf(vr->cfh,"out/source address translation on Interface %s\n",idstring);
   fhprintf(vr->cfh,"%s -> ",n2char(PKT_SOURCE(p)));
   fhprintf(vr->cfh,"%s\n",n2char(na));
#endif
      PKT_SOURCE(p)=na;
    }
    
#ifdef DEBUG_PACKETTRANSMISSION
   fhprintf(vr->cfh, "--- DEBUG_PACKETTRANSMISSION ---\n");
   fhprintf(vr->cfh,"outgoing on interface %s\n",idstring);
   fhprintf(vr->cfh,"%s\n",pkt2char(p));
#endif

   //
   // send the packet
   //

    
    ++tx_stat;
    if(!sender){
      write(fh_w, p, PKT_LENGTH(p));
      return 0;
    }
    sender->sendbuff((char*)p,PKT_LENGTH(p));
  }

  return 0;
}

//
// this function is called, when somebody wants to send
// a packet over this interface
//

void Interface::send_packet(Packet *p)
{
  
#ifdef DEBUG_PACKETTRANSMISSION
  fhprintf(vr->cfh, "--- DEBUG_PACKETTRANSMISSION ---\n");
  fhprintf(vr->cfh, "enqueing packet on Interface %s\n",idstring);
  fhprintf(vr->cfh,"%s\n", pkt2char(p));
#endif
  
  // ipip tunnel

  if(connectiontype==ipiptunnel){
    ipip_encap(p);
    return;
  }


  //
  // put packet to queueing system
  //
  
  q_send->root_enque(p);
  
  
  //
  // activate write event
  //

  event->suspended=0;		// start event scheduling
  return;
}


int Interface::ipip_encap(Packet *p)
{
  char buffer[2048]; 
  Packet *n=(Packet*)buffer;

#ifdef DEBUG_PACKETTRANSMISSION
  fhprintf(vr->cfh,"DEBUG_INF %s IPIP %s",idstring ,pkt2char(p));
#endif

  // no packet split
  
  if(PKT_LENGTH(p)+20<=1500){
    memset(buffer,0,2048);
    memcpy(buffer+20,p,PKT_LENGTH(p));
    n->ihl=5;
    n->version=4;
    n->tos=0;
    n->prot=4;			// ipip tunnel
    n->len=htons(PKT_LENGTH(p)+20);
    n->source=ip;
    n->dest=ipip_destination;
    n->ttl=128;
    n->id=packet_ids++;
    n->frag=0;
    n->crc=0;
    vr->routing->route(n);
    return 0;
  }

  // split packet in 800 byte und 700 byte payload

  memset(buffer,0,2048);
  memcpy(buffer+20,p,800);
  n->ihl=5;
  n->version=4;
  n->tos=0;
  n->prot=4;			// ipip tunnel
  n->len=htons(820);
  n->source=ip;
  n->dest=ipip_destination;
  n->ttl=128;
  n->id=packet_ids;
  n->frag=0x20;	// MF bit is set, fragment ist 0
  n->crc=0;
  vr->routing->route(n); // packet schicken

  int l=PKT_LENGTH(p)-800;
  memset(buffer,0,2048);
  memcpy(buffer+20,p,l);
  n->ihl=5;
  n->version=4;
  n->tos=0;
  n->prot=4;			// ipip tunnel
  n->len=htons(l+20);
  n->source=ip;
  n->dest=ipip_destination;
  n->ttl=128;
  n->id=packet_ids;
  n->frag=htons(100);	// MF bit + 100 x 8 byte fragmentlaenge
  n->crc=0;
  vr->routing->route(n);

  packet_ids++;

  return 0;
}


int Interface::Setup_IPIPtunnel(unsigned long dest_addr, unsigned long mtu)
{
  
  if(connectiontype!=disconnected){
    return 1;
  }

  connectiontype=ipiptunnel;
  ipip_mtu=mtu;
  ipip_destination=dest_addr;

  return 0;
}


