#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdio.h> 
#include "ipnmx.H"
#include "tools.H"

IPNMx::IPNMx(int n)
{
  nof_allocated=n;
  nof_entries=0;
  table=new IPNMxRecord*[nof_allocated];
  memset(table,0,sizeof(IPNMxRecord*)*nof_allocated);
  return;
}

IPNMx::~IPNMx()
{
  for(int i=0;i<nof_allocated;++i){
    delete table[i];
  }
  return;
}

IPNMxRecord* IPNMx::lookup(Packet *p)
{
  int h=0;
  IPNMxRecord *v=NULL;

  for(int i=0;i<nof_entries;++i){
    int k=table[i]->hits(p);
    if(k>h){
      h=k; 
      v=table[i];
    }
  }
  return v;
}

int IPNMx::add(IPNMxRecord& m)
{

  if(nof_entries+1>nof_allocated)return 1;
  table[nof_entries]=new IPNMxRecord(m);
  table[nof_entries]->init();
  nof_entries++;
  sort();
  return 0;
}

int IPNMx::del(IPNMxRecord &m)
{
  int i;
  for(i=0;i<nof_entries;++i){
    if(table[i]->equals(&m)){
      delete table[i];
      table[i]=NULL;
      break;
    }
  }
  if(i==nof_entries) return 1;
  cleanup();
  sort();
  return 0;
}

int IPNMx::get_first(IPNMxRecord &m)
{
  list_counter=0;

  if(!nof_entries || !table[list_counter]){return 0;}
  // printf("first: %s \n",n2char(table[list_counter]->sip_add));
  memcpy(&m,table[list_counter],sizeof(IPNMxRecord));
  list_counter++;
  return 1;
}

int IPNMx::get_next(IPNMxRecord &m)
{
  if(!table[list_counter] || list_counter>=nof_entries)return 0;
  //  printf("next: %s \n",n2char(table[list_counter]->sip_add));
  memcpy(&m,table[list_counter],sizeof(IPNMxRecord));
  list_counter++;
  return 1;
}

void IPNMx::sort()
{
  int m;
  do{
    m=0;
    for(int i=0;i<nof_entries-1;++i){
      if(table[i+1]->gt(table[i])){
	IPNMxRecord *r;
	r=table[i];
	table[i]=table[i+1];
	table[i+1]=r;
	++m;
      }
    }
  }while(m);
  return;
}

void IPNMx::cleanup()
{
  int i,j;
  for(i=j=0;i<nof_entries;++i){
    if(table[i]!=NULL)
      table[j++]=table[i];
  }
  nof_entries=j;
  return;
}

// ---------------------------------------------------------------------------

bool IPNMxRecord::equals(IPNMxRecord *r)
{
  return !(bool)bcmp(this,r,sizeof(IPNMxRecord));
}

bool IPNMxRecord::gt(IPNMxRecord *r)
{
  if(this->dip_add > r->dip_add)return 1;
  if(this->dip_add < r->dip_add)return 0;
  if(this->dnm_num > r->dnm_num)return 1;
  if(this->dnm_num < r->dnm_num)return 0;

  if(this->sip_add > r->sip_add)return 1;
  if(this->sip_add < r->sip_add)return 0;
  if(this->snm_num > r->snm_num)return 1;
  if(this->snm_num < r->snm_num)return 0;

  if(this->proto > r->proto)return 1;
  if(this->proto < r->proto)return 0;

  if(this->tos > r->tos)return 1;
  if(this->tos < r->tos)return 0;

  return 0;
}


bool IPNMxRecord::lt(IPNMxRecord *r)
{
  return !gt(r);
}

int IPNMxRecord::hits(Packet *p)
{
  int k=1;
  if(dip_add){
    if((PKT_DEST(p)&dnm_add) != (dip_add&dnm_add))return 0;
    k+=dnm_num+1;
  }
  if(sip_add){
    if((PKT_SOURCE(p)&snm_add) != (sip_add&snm_add))return 0;
    k+=snm_num+1;
  }
  if(proto){
    if(proto!=PKT_PROTO(p))return 0;
    k+=1;
  }
  if(tos){
    if(tos!=PKT_TOS(p))return 0;
    k+=1;
  }
  return k;
}


IPNMxRecord::IPNMxRecord(Packet *p)
{
  sip_add=PKT_SOURCE(p);
  snm_add=0xffffffff;
  dip_add=PKT_DEST(p);
  dnm_add=0xffffffff;
  proto=PKT_PROTO(p);
  tos=PKT_TOS(p);
  content=NULL;
  return;
}

IPNMxRecord::IPNMxRecord(unsigned long ip, unsigned long nm, void *v)
{
  memset(this,0,sizeof(IPNMxRecord));
  dip_add=ip;
  dnm_add=nm;
  dnm_num=nm2n(nm);
  content=v;
  return;
}

IPNMxRecord::IPNMxRecord(unsigned long sip, unsigned long snm, 
			 unsigned long dip, unsigned long dnm,
			 unsigned short p, unsigned short t, void *v)
{
  memset(this,0,sizeof(IPNMxRecord));
  dip_add=dip;
  dnm_add=dnm;
  dnm_num=nm2n(dnm);
  sip_add=sip;
  snm_add=snm;
  snm_num=nm2n(snm);
  content=v;
  return;
}

int IPNMxRecord::set(char *d, char *s, int p, int t, void *v)
{
  memset(this,0,sizeof(IPNMxRecord));
  char2ipnm(s, &sip_add, &snm_add);
  char2ipnm(d, &dip_add, &dnm_add);
  dnm_num=nm2n(dnm_add);
  snm_num=nm2n(snm_add);
  proto=p;
  tos=t;
  content=v;
  return 0;
}


IPNMxRecord::IPNMxRecord(char *d, char *s, int p, int t, void *v)
{
  memset(this,0,sizeof(IPNMxRecord));
  char2ipnm(s, &sip_add, &snm_add);
  char2ipnm(d, &dip_add, &dnm_add);
  dnm_num=nm2n(dnm_add);
  snm_num=nm2n(snm_add);
  proto=p;
  tos=t;
  content=v;
  return;
}



void IPNMxRecord::clear()
{
  memset(this,0,sizeof(IPNMxRecord));
  return;
}

void IPNMxRecord::init()
{
  snm_num=nm2n(snm_add);
  dnm_num=nm2n(dnm_add);
  return;
}


IPNMxRecord::IPNMxRecord()
{
  sip_add=0; snm_add=0; snm_num=0;
  dip_add=0; dnm_add=0; dnm_num=0;
  proto=0; tos=0;
  content=NULL;
  return;
}

IPNMxRecord::~IPNMxRecord()
{
  return;
}

/*
CON & operator << (CON &os, IPNMxRecord *r)
{
  if(!r){
    os << "null";
    return os;
  }
  char buff1[256], buff2[256], buff3[256];
  char *ps;
  if(!r){ 
    os << "null" << "\n";
    return os;
  }
  sprintf(buff1,"%s/%d",n2char(r->dip_add), r->dnm_num);
  sprintf(buff2,"%s/%d",n2char(r->sip_add), r->snm_num);
  switch(r->proto)
    {
    case 0 : ps="none"; break;
    case 6 : ps="TCP";  break;
    case 17: ps="UDP";  break;

    default: {
      ps="number"; sprintf(ps,"%d",r->proto);
    }
    }
  
  sprintf(buff3,"%-20s %-20s %-4s %-4d", buff1, buff2, ps, r->tos);
  os << buff3;
  return os;
}



*/







