#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "qco.H"
#include "qcclassifier.H"
#include "ipnmx.H"
#include "tools.H"
 
QCclassifier::QCclassifier(QCroot *q) : QCO(q)
{
  max_nof_ins=1; 
  ins=new QCO*[max_nof_ins];
  memset(ins,0,sizeof(QCO*)*max_nof_ins);
  max_nof_outs=32;
  outs=new QCO*[max_nof_outs];
  memset(outs,0,sizeof(QCO*)*max_nof_outs);
  ipnm=new IPNMx(256);
}

QCclassifier::~QCclassifier()
{
  delete outs;
  delete ins;
  delete ipnm;
}

Packet *QCclassifier::nextdeque(QCO *target)
{
  if(!ins)return NULL;

  Packet *p=ins[0]->nextdeque(this);
  if(!p)return NULL;

  IPNMxRecord *r=ipnm->lookup(p);
  if(!r){
    fhprintf(vr->cfh,"qcclassifier: no valid class found, dropping packet\n");
    return NULL;
  }
  if(outs[(int)r->content]==target)return p;

  return NULL;
}

Packet * QCclassifier::deque(QCO *target)
{

  if(!ins) return NULL;

  Packet *p=ins[0]->nextdeque(this);
  if(!p)return NULL;

  IPNMxRecord *r=ipnm->lookup(p);
  if(!r){
    fhprintf(vr->cfh,"qcclassifier: no valid class found, dropping packet\n");
    return NULL;
  }
  if(outs[(int)r->content]==target)return deque(this);
  

#ifdef DEBUG_QUEUEING
  fhputs("--- deque ---\n",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  return NULL;
}

int QCclassifier::enque(Packet *p)
{

#ifdef DEBUG_QUEUEING
  fhputs("--- enque ---",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  if(!outs)return QUEUE_NAK;
  IPNMxRecord *r=ipnm->lookup(p);
  if(!r){
    fhprintf(vr->cfh,"qcclassifier: no valid class found, dropping packet\n");
    return QUEUE_ACK;
  }
  if(!outs[(int)r->content]){
    fhprintf(vr->cfh,"qcclassifier: irregular classifier\n");
    return QUEUE_NAK;
  }
  return outs[(int)r->content]->enque(p);
}

VRRB *QCclassifier::APIconfigure(VRCB *v)
{
  IF_QS_VRCB *cb=(IF_QS_VRCB*)v;
  switch(cb->component_command)
    {
    case 1:		// reset counters
      {
	VRRB *rb=(VRRB*)malloc(sizeof(VRRB));
	rb->handle=v->handle;
	rb->len=sizeof(VRRB);

	/*
	 * reset counter code
	 */

	return rb;
      }

    case 2:		// return statistics
      {
	QS_CLASSIFIER_STATISTICS_VRRB *rb=(QS_CLASSIFIER_STATISTICS_VRRB*)
	  malloc(sizeof(QS_CLASSIFIER_STATISTICS_VRRB)+nof_ins*sizeof(GEN_CLASSIFIER_RECORD));
	
	rb->len=sizeof(QS_CLASSIFIER_STATISTICS_VRRB)+ipnm->size()*sizeof(GEN_CLASSIFIER_RECORD);
	rb->handle=cb->handle;
	rb->component_type=type_id(); 
	rb->component_id=get_id(this);
	rb->enques=0;
	rb->deques=0;
	
	GEN_CLASSIFIER_RECORD *rec=(GEN_CLASSIFIER_RECORD *)(rb+1); // append to VRRB 
	
	IPNMxRecord a;
	int r=ipnm->get_first(a);
	int i=0;
	while(r){
	  rec[i].source_ip=a.sip_add;
	  rec[i].source_nm=a.snm_add;
	  rec[i].dest_ip=a.dip_add;	
	  rec[i].dest_nm=a.dnm_add;
	  rec[i].tos=a.tos;		
	  rec[i].proto=a.proto;	
	  int k=(int)a.content;
	  rec[i].type=outs[k]->type_id();
	  rec[i].id=get_id(outs[k]);
	  ++i;
	  r=ipnm->get_next(a);
	}
	rb->len=sizeof(QS_CLASSIFIER_STATISTICS_VRRB)+i*sizeof(GEN_CLASSIFIER_RECORD);
	return (VRRB*)rb;
      }      
      
    case 511:
      {
	struct dummy0 : public IF_QS_VRCB
	{
	  unsigned short comp_id;
	  unsigned long s_add;
	  unsigned long s_add_nm;
	  unsigned long d_add;
	  unsigned long d_add_nm;
	  unsigned short proto;
	  unsigned short tos;
	};
	
	struct dummy1 : public QS_CLASSIFIER_STATISTICS_VRRB
	{
	  GEN_CLASSIFIER_RECORD rec;
	};
	
	struct dummy0 *cb=(struct dummy0*)v;
	struct dummy1 *rb=(struct dummy1*)malloc(sizeof(struct dummy1));
	
	rb->handle=cb->handle;
	rb->len=sizeof(struct dummy1);
	rb->component_type=type_id(); 
	rb->component_id=get_id(this);
	memset(&(rb->rec),0,sizeof(GEN_CLASSIFIER_RECORD));
	
	int j, idx=-1;
	for(j=0;j<nof_outs;++j){
	  if(get_id(outs[j])==cb->comp_id){
	    idx=j;	 break;
	  }
	}
	if(idx==-1){
	  return (VRRB*)rb;
	}
	
	IPNMxRecord a(cb->s_add, cb->s_add_nm, cb->d_add, cb->d_add_nm, cb->proto, cb->tos, (void*)idx);
	ipnm->add(a);
	
	rb->rec.type=outs[idx]->type_id();
	rb->rec.id=cb->comp_id;
	rb->rec.source_ip=cb->s_add;
	rb->rec.source_nm=cb->s_add_nm;
	rb->rec.dest_ip=cb->d_add;
	rb->rec.dest_nm=cb->d_add_nm;
	rb->rec.tos=cb->tos;
	rb->rec.proto=cb->proto;
	
	return (VRRB*)rb;
      }
    }

  return NULL;
}

/*

int QCclassifier::del(int argc, char **argv, int i)
{
  int proto=atoi(argv[i+2]);
  int tos=atoi(argv[i+3]);
  char *s=argv[i];
  char *d=argv[i];

  if(!strcmp(argv[i+2],"tcp"))proto=6;
  if(!strcmp(argv[i+2],"udp"))proto=17;
  if(!strcmp(argv[i+2],"none"))proto=0;
  if(!strcmp(argv[i+2],"-"))proto=0;

  if(!strcmp(argv[i+3],"none"))tos=0;
  if(!strcmp(argv[i+3],"-"))tos=0;

  if(!strcmp(argv[i],"none"))d="0.0.0.0/0";
  if(!strcmp(argv[i],"-"))d="0.0.0.0/0";

  if(!strcmp(d,"none"))d="0.0.0.0/0";
  if(!strcmp(s,"-"))s="0.0.0.0/0";


  IPNMxRecord a(d, s, proto, tos, NULL);

  if(ipnm->del(a))
    fhputs("cannot delete entry\n",vr->cfh);
  else
    fhputs("entry deleted\n",vr->cfh);
  return 0;
}
*/





