#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "tools.H"
#include "qcscheduler.H"
#include "qco.H"

QCscheduler::QCscheduler(QCroot *q) : QCO(q)
{
  mode=SCHED_PRR;
  rangen_seed=42;
  ins_choosen=-1;
  rr_cnt=0;

  max_nof_ins=32; 
  ins=new QCO*[max_nof_ins];
  memset(ins,0,sizeof(QCO*)*max_nof_ins);

  weights=new unsigned short[max_nof_ins];
  for(int i=0; i<max_nof_ins;i++)weights[max_nof_ins-i-1]=i;

  dehits=new unsigned long[max_nof_ins];
  memset(dehits,0,sizeof(unsigned long)*max_nof_ins);

  max_nof_outs=1;
  outs=new QCO*[max_nof_outs];
  memset(outs,0,sizeof(QCO*)*max_nof_outs);

  idx_avail=new int[max_nof_ins];
  memset(idx_avail,0,sizeof(int)*max_nof_ins);

  pkt_avail=new Packet*[max_nof_ins];
  memset(pkt_avail,0,sizeof(Packet*)*max_nof_ins);

  return;
}

QCscheduler::~QCscheduler()
{
  delete idx_avail;
  delete pkt_avail;
  delete dehits;
  delete weights;
  delete outs;
  delete ins;
}

Packet * QCscheduler::deque(QCO *x)
{
  if(!ins)return NULL;
  
#ifdef DEBUG_QUEUEING
  fhputs("--- deque ---\n",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  if(mode==SCHED_RR){
    Packet *p;
    int rr_cnt_old=rr_cnt;
    do{
      p=ins[rr_cnt]->deque(this);
      if(p)return p;
      ++rr_cnt;
      if(rr_cnt>=nof_ins)rr_cnt=0;
    }while(rr_cnt!=rr_cnt_old);
    return NULL;
  }
  
  if(mode==SCHED_PRR){
    for(int i=0;i<nof_ins;++i){
      Packet *p=ins[i]->deque(this);
      if(p){
	++dehits[i];
	return p;
      }
    }
    return NULL;
  }
  
  if(mode==SCHED_WFQ){
    Packet *p;
    if(ins_choosen==-1){
      p=nextdeque(this);
      if(!p)return NULL;
    }
    
    p=ins[ins_choosen]->deque(this);
    ins_choosen=-1;
    //    ++dehits[i];
    return p;
  }
  return NULL;
}

Packet * QCscheduler::nextdeque(QCO *x)
{
  if(mode==SCHED_RR){
    Packet *p;
    int rr_cnt_old=rr_cnt;
    do{
      if(!ins[rr_cnt]){
	fhprintf(vr->cfh,"fatal !! %d\n",rr_cnt);
	exit(0);
      }
      p=ins[rr_cnt]->nextdeque(this);
      if(p)return p;
      ++rr_cnt;
      if(rr_cnt>=nof_ins)rr_cnt=0;
    }while(rr_cnt!=rr_cnt_old);
    return NULL;
  }
    
  if(mode==SCHED_PRR){
    for(int i=0;i<nof_ins;++i){
      Packet *p=ins[i]->nextdeque(this);
      if(p){
	return p;
      }
    }
    return NULL;
  }

  if(mode==SCHED_WFQ){
    int k=0; 
    unsigned long w=0;
    ins_choosen=-1;

    for(int i=0;i<nof_ins;++i){
      Packet *p=ins[i]->nextdeque(this);
      if(p){
	pkt_avail[k]=p;
	idx_avail[k]=i;
	++k;
	w+=weights[i];
      }
    }
    if(!k)return NULL;

    long r=w*rangen()/0xffff;
    for(int i=0;i<k;++i){
      r-=weights[idx_avail[i]];
      if(r<=0){
	ins_choosen=idx_avail[i];
	return pkt_avail[i];
      }
    }
    fhputs("error\n",vr->cfh);
    return NULL;
  }
  return NULL;
}


int QCscheduler::enque(Packet *p)
{
#ifdef DEBUG_QUEUEING
  fhputs("--- enque ---",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  return QUEUE_NAK;

  if(!outs)return 1;
  return outs[0]->enque(p);
}



VRRB *QCscheduler::APIconfigure(VRCB *v)
{
  IF_QS_VRCB *cb=(IF_QS_VRCB*)v;

  switch(cb->component_command)
    {
    case 1:		// reset counters
      {
	VRRB *rb=(VRRB*)malloc(sizeof(VRRB));
	rb->handle=v->handle;
	rb->len=sizeof(VRRB);

	memset(dehits,0,sizeof(unsigned long)*max_nof_outs);

	return rb;
      }

    case 2:		// return statistics
      {
	QS_SCHEDULER_STATISTICS_VRRB *rb=(QS_SCHEDULER_STATISTICS_VRRB*)
	  malloc(sizeof(QS_SCHEDULER_STATISTICS_VRRB)+nof_ins*sizeof(GEN_SCHEDULER_RECORD));
	
	rb->len=sizeof(QS_SCHEDULER_STATISTICS_VRRB)+nof_ins*sizeof(GEN_SCHEDULER_RECORD);
	rb->handle=cb->handle;
	rb->component_type=type_id(); 
	rb->component_id=get_id(this);
	rb->mode=mode;
	
	GEN_SCHEDULER_RECORD *rec=(GEN_SCHEDULER_RECORD *)(rb+1); // append to VRRB 
	
	for(int i=0; i< nof_ins;++i){
	  if(ins[i]){
	    rec[i].type=ins[i]->type_id();
	    rec[i].id=get_id(ins[i]);
	    rec[i].misc=weights[i];          // priority, order 
	    rec[i].deque_hits=0;
	  }
	}
	
	return (VRRB*)rb;
      }
    case 411:			// change weight / order
      {
	struct dummy0 : public IF_QS_VRCB
	{
	  unsigned short comp_id;
	  unsigned short misc;
	};
	struct dummy0 *cb=(struct dummy0*)v;
	
	int i;
	for(i=0; i<nof_ins;++i){
	  if(ins[i])
	    if(get_id(ins[i])==cb->comp_id)break;
	}
	struct dummy1 : public VRRB
	{
	  GEN_SCHEDULER_RECORD rec;
	};
	struct dummy1 *rb=(struct dummy1*)malloc(sizeof(struct dummy1));
	rb->handle=cb->handle;
	if(i!=nof_ins){	// weight changed
	  weights[i]=cb->misc;
	  rb->len=sizeof(struct dummy1);
	  rb->rec.type=ins[i]->type_id();
	  rb->rec.id=get_id(ins[i]);
	  rb->rec.misc=cb->misc;
	  rb->rec.deque_hits=0;
	  ins_sort();		// update the order of the comps
	  return (VRRB*)rb;
	}
	
	rb->len=sizeof(struct VRRB); // error
	return (VRRB*)rb;

      }
    case 410:			// change mode
      {
	struct dummy0 : public IF_QS_VRCB
	{
	  unsigned short new_sched_mode;
	};
	struct dummy0 *cb=(struct dummy0*)v;
	mode=cb->new_sched_mode;

	QS_SCHEDULER_STATISTICS_VRRB *rb=(QS_SCHEDULER_STATISTICS_VRRB*)
	  malloc(sizeof(QS_SCHEDULER_STATISTICS_VRRB));
	rb->len=sizeof(sizeof(QS_SCHEDULER_STATISTICS_VRRB));
	rb->handle=cb->handle;
	rb->component_type=type_id(); 
	rb->component_id=get_id(this);
	rb->mode=mode;
	return (VRRB*)rb;
	break;
      }

    default:			// some error occured
      {
	break;
      }
    }

  return NULL;
}

/*
 * sorts the incoming components according to their weights
 */

void QCscheduler::ins_sort()
{
  int dirty;
  do{
    dirty=0;
    for(int i=0;i<nof_ins-1;++i){
      if(weights[i]<weights[i+1]){
	QCO *q; 
	int w;
	unsigned long h;
	q=ins[i]; w=weights[i]; h=dehits[i];
	ins[i]    =ins[i+1]; 
	weights[i]=weights[i+1]; 
        dehits[i] =dehits[i+1];

	ins[i+1]=q; 
	weights[i+1]=w; 
	dehits[i+1]=h;

	dirty=1;
      }
    }
  }while(dirty);
  return;
}

/*
 * random number generation
 */

unsigned short QCscheduler::rangen()
{
  rangen_seed = rangen_seed * 1664525 + 1013904223;
  return (unsigned short)abs(rangen_seed%0xffff);
}



