#ifndef _shell_H_
#define _shell_H_

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>


class Shell : public virtual API
{
public:
  Shell(VR*);
  ~Shell();
  void interprete(int,char**,char *,int);	//- interpretes command string and returns result
  int interrupt(VRevent *e);	//- is called by the Event scheduler 
  int send_RB(VRRB *);		//- send an asynchronous RB to the API 
  int forward(FFilter *, Packet *p);

private:

  void interprete_sys(int, char**, char*,int); //- interpretation of system commands 
  void interprete_ifconfig(int,char**,char*,int); //- interpretes the ifconfig command syntax
  void interprete_route(int,char**,char*,int); //- interpretes the route command syntax
  void interprete_load(int,char**,char*,int); //- interpretes the load command
  void interprete_test(int,char**,char*,int); //- interpretes the load command
  void interprete_attach(int argc, char **argv, char *s, int l);
 
  int TCPopen();		//- open TCP connection 
  int TCPpoll();		//- poll for new connection
  int TCPclose();		//- close connection


  int dsmarker_rule_set(int argc, char **argv, int if_no, int comp_id, char *s);
  int dsmarker_rule_erase(int argc, char **argv, int if_no, int comp_id, char *s);
  int parse_dsmarkerrule(int argc, char **argv, DS_MARKER_RECORD *dsr);
  int trio_configure(int argc, char **argv, int if_no, int comp_id, char *s);
  int droptail_configure(int argc, char **argv, int if_no, int comp_id, char *s);
  int tbf_configure(int argc, char **argv, int if_no, int comp_id, char *s);
  int scheduler_wt_setting(int argc, char **argv, int if_no, int comp_id, char *s);
  int scheduler_mode_setting(int argc, char **argv, int if_no, int comp_id, char *s);
 
  IF_INFORMATION_RB *get_interface_by_name(char *);

  int is_ifname(char *);
  int is_ipadd(char *s);
  
  void banner();
  int password_check(char *);
  char *lobmode2string(unsigned short mode);
  char *qid2string(int);
  char *filterrec2string(FILTER_REC *f);
  int IF_INFORMATION_RB2string(IF_INFORMATION_RB *, char*, int l=0);
  int EVENT_REC2string(struct EVENT_REC *ev, char *s, int l=0);
  int split(char *s, char **argv);
  int is_delimiter(char c);
  char hostname[256];
  int sfh,sock;			//- open socket fh and socket fh
  struct sockaddr_in addr, raddr;
  struct hostent *hent;
  int raddr_len;
  int port;
  int binary_mode;
  int pw_prompt;
};


#endif
