#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "config.H"
#include "vr.H"
#include "packet.H"
#include "forwarder_if.H"
#include "socks.H"
#include "tools.H"
#include "forwarder.H"
#include "routing.H"
#include "interface.H"

VR *Sock::vr=NULL;
int RawSock::number_of_sockets=0;
RawSock* RawSock::socks[MAX_NUMBER_OF_RAW_SOCKETS];
int UDPSock::number_of_sockets=0;
UDPSock* UDPSock::socks[MAX_NUMBER_OF_UDP_SOCKETS];


Sock::Sock()
{
  return;
}



Sock::Sock(VR *v)
{
  vr=v;
  return;
}



int Sock::transmit_packet(char *buffer, int n)
{
  return ::write(vr->pf->hostif[1],buffer,n);
}



Sock::~Sock()
{
  return;
}

RawSock::RawSock(VR *v) : Sock(v)
{
#ifdef DEBUG_RAWSOCKETS
  fhputs("DEBUG_RAWSOCKETS: is parent socket",vr->cfh);
#endif

  forwarder=NULL;
  mode=VRSOCK_PARENT;
  number_of_sockets=0;
  memset(socks,0,sizeof(Sock*)*MAX_NUMBER_OF_RAW_SOCKETS);
}



RawSock::RawSock(int m) // proto: icmp, mode:1 read, mode:2 write
{
  // packete werden von VR nach sv[0] geschrieben
  // und von applications von sv[1] gelesen 
  forwarder=NULL;
  mode=m;
  proto=0;
  bytes_in_socket=0;
  hdr_include=0;
  blocking=1;
  d_addr=0;
  s_addr=0;
  ttl=0;
  
#ifdef DEBUG_RAWSOCKETS
  fhprintf(vr->cfh,"DEBUG_RAWSOCKETS: raw socket created mode: %d\n",mode);
#endif

  if(mode&VRSOCK_PARENT){		// This is the parent
#ifdef DEBUG_RAWSOCKETS
  fhputs("DEBUG_RAWSOCKETS: is parent socket",vr->cfh);
#endif
    number_of_sockets=0;
    memset(socks,0,sizeof(Sock*)*MAX_NUMBER_OF_RAW_SOCKETS);
    return;
  }
 
  if(number_of_sockets>=MAX_NUMBER_OF_RAW_SOCKETS){
#ifdef DEBUG_RAWSOCKETS
  puts("DEBUG_RAWSOCKETS: Error too many sockets");
#endif
    mode|=VRSOCK_DISABLED;
    return;
  }

  number_of_sockets++;
  for(int i=0;i<MAX_NUMBER_OF_RAW_SOCKETS;++i)
    if(socks[i]==NULL){
      socks[i]=this;
      socks_index=i;
      break;
    }
#ifdef DEBUG_RAWSOCKETS
  printf("DEBUG_RAWSOCKETS: index is %d\n",socks_index);
#endif

  if(socketpair(AF_UNIX, SOCK_STREAM, 0, sv)){
    mode|=VRSOCK_DISABLED;
#ifdef DEBUG_RAWSOCKETS
  puts("DEBUG_RAWSOCKETS: Error creating socketpair\n");
#endif
    return;
  }
}



int RawSock::dispatch(Packet *p)
{
#ifdef DEBUG_RAWSOCKETS
  fhprintf(vr->cfh,"--- DEBUG_RAWSOCKETS ---\n");
  fhprintf(vr->cfh,"raw socket dispatcher\n");
  fhprintf(vr->cfh,"packet %s\n",pkt2char(p));
#endif
    
  if((mode&VRSOCK_PARENT)==0) return VRSOCK_ERROR;

  for(int i=0;i<MAX_NUMBER_OF_RAW_SOCKETS;++i){
    int l=0;
    if(!socks[i])continue;	// no such socket
    if((socks[i]->mode&VRSOCK_READ)==0)continue; // socket does not read
    if(socks[i]->proto && socks[i]->proto!=PKT_PROTO(p))continue; // wrong protocol

    if(socks[i]->hdr_include){
      if(socks[i]->forwarder){
	socks[i]->forwarder->forward(this,p);
	l+=PKT_LENGTH(p);
      } else {
	l=::write(socks[i]->sv[0], p, PKT_LENGTH(p));
      }
    } else {
      l=::write(socks[i]->sv[0], PKT_DATA(p), PKT_DATALEN(p));
    }

    socks[i]->bytes_in_socket+=l;
  }
  
  return PKT_LENGTH(p);
}



int RawSock::write(char *buffer, int n)
{
#ifdef DEBUG_RAWSOCKETS
  fhprintf(vr->cfh,"--- DEBUG_RAWSOCKETS ---\n");
  fhprintf(vr->cfh,"wrting packet\n");
#endif

  if(mode&VRSOCK_PARENT)return VRSOCK_ERROR;
  if(mode&VRSOCK_DISABLED)return VRSOCK_ERROR;
  if((mode&VRSOCK_WRITE)==0)return VRSOCK_BADMODE;

  if(hdr_include)return transmit_packet(buffer,n);

  if(!d_addr)return VRSOCK_MISSOPT;

  Packet *p= (Packet*)malloc(n+256);

  if(!ttl)ttl=vr->ttl_default;

  memset(p,0,n+256);
  p->version=4;
  p->ihl=5;
  p->len=htons(20+n);
  p->ttl=ttl;
  p->id=vr->pf->get_packet_id();
  p->prot=proto;
  p->dest=d_addr;

  if(!s_addr){
    p->source=0;
    Interface *ifa=vr->routing->if_lookup(p);
    if(!ifa){
      for(int i=0;i<vr->num_of_interfaces;++i)
	if(vr->in[i])ifa=vr->in[i];
    }
    s_addr=ifa->get_ip();
  }
  p->source=s_addr; 

  memcpy(PKT_DATA(p),buffer,n);
  p->crc=0;
  p->crc=cksum((char*)p,PKT_HDRLEN(p));

  transmit_packet((char*)p,PKT_LENGTH(p));

  free(p);
  return n+20;
}




int RawSock::read(char *buffer, int n)
{
  if(mode&VRSOCK_PARENT)return VRSOCK_ERROR;
  if(mode&VRSOCK_DISABLED)return VRSOCK_ERROR;
  if((mode&VRSOCK_READ)==0)return VRSOCK_BADMODE;

  if(!blocking &&  !bytes_in_socket)return 0;

  int l=::read(sv[1], buffer, n);

  bytes_in_socket-=l;
  return l;
}





int RawSock::setopt(int o, unsigned long v)
{
  switch(o)
    {
    case VRSOCK_PROTO:
      proto=v;
      return 0;
    case VRSOCK_DEST:
      d_addr=v;
      return 0;
    case VRSOCK_HDRINC:
      hdr_include=v;
      return 0;
    case VRSOCK_NONBLOCK:
      blocking=v;
      return 0;
    case VRSOCK_SOURCE:
      s_addr=v;
      return 0;
    }
  return -1;
}

int RawSock::setopt(int o, ForwarderIF *v)
{
  switch(o)
    {
    case VRSOCK_DIRECT:
      forwarder=v;
      return 0;
    }
  return -1;
}



RawSock::~RawSock()
{
  if((mode&VRSOCK_DISABLED)==0 && (mode&VRSOCK_PARENT)==0){
    close(sv[0]);
    close(sv[1]);
    socks[socks_index]=NULL;
    number_of_sockets--;
  }
  return;
}




UDPSock::UDPSock(VR *v)
{
#ifdef DEBUG_UDPSOCKETS
  fhputs("DEBUG_UDPSOCKETS: is parent socket",vr->cfh);
#endif
  mode=VRSOCK_PARENT;
  number_of_sockets=0;
  memset(socks,0,sizeof(Sock*)*MAX_NUMBER_OF_UDP_SOCKETS);
  forwarder=NULL;
  return;
}



UDPSock::UDPSock(int m)
{
  // packete werden von VR nach sv[0] geschrieben
  // und von applications von sv[1] gelesen 
  mode=m;
  bytes_in_socket=0;
  hdr_include=0;
  blocking=1;
  d_addr=0;
  s_addr=0;
  ttl=0;
  forwarder=NULL;
  
#ifdef DEBUG_UDPSOCKETS
  fhprintf(vr->cfh,"DEBUG_UDPSOCKETS: UDP socket created mode: %d\n",mode);
#endif
  
 
  if(number_of_sockets>=MAX_NUMBER_OF_UDP_SOCKETS){
#ifdef DEBUG_UDPSOCKETS
    puts("DEBUG_UDPSOCKETS: Error too many UDP sockets");
#endif
    mode|=VRSOCK_DISABLED;
    return;
  }

  number_of_sockets++;
  for(int i=0;i<MAX_NUMBER_OF_UDP_SOCKETS;++i)
    if(socks[i]==NULL){
      socks[i]=this;
      socks_index=i;
      break;
    }
#ifdef DEBUG_UDPSOCKETS
  printf("DEBUG_UDPSOCKETS: index is %d\n",socks_index);
#endif
  
  if(socketpair(AF_UNIX, SOCK_STREAM, 0, sv)){
    mode|=VRSOCK_DISABLED;
#ifdef DEBUG_UDPSOCKETS
    puts("DEBUG_UDPSOCKETS: Error creating socketpair\n");
#endif
    return;
  }
}



UDPSock::~UDPSock()
{
  if((mode&VRSOCK_DISABLED)==0 && (mode&VRSOCK_PARENT)==0){
    close(sv[0]);
    close(sv[1]);
    socks[socks_index]=NULL;
    number_of_sockets--;
  }
  return;
}



int UDPSock::setopt(int o, unsigned long v)
{
  switch(o)
    {
    case VRSOCK_PROTO:
      return -1;
    case VRSOCK_DEST:
      d_addr=v;
      return 0;
    case VRSOCK_HDRINC:
      hdr_include=v;
      return 0;
    case VRSOCK_NONBLOCK:
      blocking=v;
      return 0;
    case VRSOCK_SOURCE:
      s_addr=v;
      return 0;
    case VRSOCK_TTL:
      ttl=v;
      return 0;
    case VRSOCK_LOCALPORT:
      localport=v;
      return 0;
    case VRSOCK_REMOTEPORT:
      remoteport=v;
      return 0;
    }
  return -1;
}



int UDPSock::setopt(int o, ForwarderIF *v)
{
  switch(o)
    {
    case VRSOCK_DIRECT:
      forwarder=v;
      return 0;
    }
  return -1;
}



int UDPSock::write(char *buffer, int n)
{

struct UDPhdr{
  unsigned short sp;
  unsigned short dp;
  unsigned short len;
  unsigned short crc;
};


#ifdef DEBUG_UDPSOCKETS
  fhprintf(vr->cfh,"--- DEBUG_UDPSOCKETS ---\n");
  fhprintf(vr->cfh,"writing UDP packet\n");
#endif

  if(mode&VRSOCK_PARENT)return VRSOCK_ERROR;
  if(mode&VRSOCK_DISABLED)return VRSOCK_ERROR;
  if((mode&VRSOCK_WRITE)==0)return VRSOCK_BADMODE;

  //  puts("UDPSock::write here 1");
  if(!d_addr)return VRSOCK_MISSOPT;
  if(!ttl)ttl=vr->ttl_default;

  if(!localport){
    localport=42;
  }
  if(!remoteport){
    remoteport=42;
  }

  //  puts("UDPSock::write here 2");
    
  Packet *p= (Packet*)malloc(n+256);
  memset(p,0,n+256);
  p->version=4;
  p->ihl=5;
  p->ttl=ttl;
  p->id=vr->pf->get_packet_id();
  p->prot=17;
  p->dest=d_addr;

  //  puts("UDPSock::write here 3");

  if(!s_addr){
    p->source=0;
    Interface *ifa=vr->routing->if_lookup(p);
    if(!ifa){
      for(int i=0;i<vr->num_of_interfaces;++i)
	if(vr->in[i])ifa=vr->in[i];
    }
    s_addr=ifa->get_ip();
  }

  p->source=s_addr; 
  p->len=htons(PKT_HDRLEN(p)+sizeof(struct UDPhdr)+n);

  p->crc=0;
  p->crc=cksum((char*)p,PKT_HDRLEN(p));
  
  struct UDPhdr *h=(struct UDPhdr*)PKT_DATA(p);

  //  puts("UDPSock::write here 4");
  
  h->sp=ntohs(localport);
  h->dp=ntohs(remoteport);
  h->crc=0;
  h->len=htons(n);
  p->crc=cksum((char*)h,sizeof(struct UDPhdr));

  //  puts("UDPSock::write here 5");
  
  memcpy(PKT_DATA(p)+sizeof(struct UDPhdr),buffer,n);

  //  puts("UDPSock::write here 6");

#ifdef DEBUG_UDPSOCKETS
  fhprintf(vr->cfh, "packet %s\n",pkt2char(p));
#endif

  transmit_packet((char*)p,PKT_LENGTH(p));

  //  puts("UDPSock::write here 7");
  free(p);
  return n;
}



int UDPSock::read(char *buffer, int n)
{
  if(mode&VRSOCK_PARENT)return VRSOCK_ERROR;
  if(mode&VRSOCK_DISABLED)return VRSOCK_ERROR;
  if((mode&VRSOCK_READ)==0)return VRSOCK_BADMODE;
  
  if(!blocking &&  !bytes_in_socket)return 0;
  
  int l=::read(sv[1], buffer, n);
  
  bytes_in_socket-=l;
  return l;
}




int UDPSock::process(PSPacket *psp)
{
  dispatch(&psp->pkt);
  return 0;
}



int UDPSock::dispatch(Packet *p)
{
#ifdef DEBUG_UDPSOCKETS
  puts("DEBUG_UDPSOCKETS: udp socket dispatchter\n");
#endif
  
  if((mode&VRSOCK_PARENT)==0) return VRSOCK_ERROR;

  for(int i=0;i<MAX_NUMBER_OF_UDP_SOCKETS;++i){
    int l=0;
    if(!socks[i])continue;	// no such socket
    if((socks[i]->mode&VRSOCK_READ)==0)continue; // socket does not read
   
    if(socks[i]->hdr_include){
      if(socks[i]->forwarder){
	socks[i]->forwarder->forward(this,p);
	l+=PKT_LENGTH(p);
      } else {
	l=::write(socks[i]->sv[0], p, PKT_LENGTH(p));
      }
    } else {
      l=::write(socks[i]->sv[0], PKT_DATA(p), PKT_DATALEN(p));
    }

    socks[i]->bytes_in_socket+=l;
  }
  
  return PKT_LENGTH(p);
}

