#include <string.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <stdarg.h>
#include "packet.H"
#include "tools.H"
#include "config.H"

char *configlinesplit(char *s)
{
  char *p=strchr(s,':');
  if(!p)return NULL;		// invalid line
  char *q=strchr(s,'#');
  if(q){
    if(q<p)return NULL;		// 
    *q='\0';
  }
  *p='\0'; ++p;
  for(;*p && isspace(*p);++p);
  if(!*p)return NULL;
  strip_trailing(p,"\n \t\r");
  return p;
}

int istrue(char *s)
{
  if(!strcmp(s,"on"))return -1;
  if(!strcmp(s,"true"))return -1;
  if(!strcmp(s,"yes"))return -1;
  return 0;
}


int fhprintf(int fh, const char *format, ...)
{
  int r=0;
#ifdef VR_USE_VASPRINTF
  char *buffer;
  va_list ap;
  va_start(ap,format);
  r=vasprintf(&buffer,format, ap);
  write(fh, buffer, strlen(buffer));
  free(buffer);
#else

  char buffer[8192];
  va_list ap;
  va_start(ap,format);
  r=vsnprintf(buffer, 8192, format, ap);
  write(fh, buffer, strlen(buffer));
#endif
  return r;
}

int fhputs(const char *s, int fh)
{
  return write(fh, s, strlen(s));
}

char *int2bitstring(int i)
{
  static char buffer[17]; 
  int j;
  for(j=0;j<16;++j){
    if(i&1)buffer[j]='1'; else buffer[j]='0';
    i=i>>1;
  }
  buffer[j]=0;
  return buffer;
}

int getargument(char *buffer, char *key, char *dest, int maxlen)
{
  int len=strlen(buffer);
  char *p=strstr(buffer,key);
  char *q=dest; *q=0;
  if(!p)return 0;
  if(!maxlen)return 0;

  p+=strlen(key);
  if(p>=buffer+len)return 0;

  for(;*p && isspace(*p); ++p);
  for(;*p && !isspace(*p) && p-buffer<maxlen;++p){
    *q=*p; ++q;
  }
  *q=0;
  return 1;
}


int strip_trailing(char *p, char *s)
{
  if(!p || !s) return 0;
  int i=strlen(p);

  if(!i)return 0;
  --i;

  for(;i>=0;--i){
    if(strchr(s,(int)p[i])){
      p[i]=0; 
    }else {
      break;
    }
  }
  return 0;
}

int arg_query(int argc, char **argv, char p, char *option, int len)
{
  if(len)option[0]=0;

  for(int i=0;i<argc;++i){
    if((argv[i][0]=='-')){
      char *pv=strchr(argv[i],p);
      if(pv){
	if(!len)return 1;
	if((*(pv+1)=='\0') && i<argc-1 && argv[i+1][0]!='-'){
	  strncpy(option, argv[i+1],len);
	  return 1;
	}
      }
    }
  }
  return 0;
}


int isnumber(char *s)
{
  for(char* c=s; *c; ++c)if(!isdigit(*c))return 0;
  return -1;
}

unsigned short cksum(char *ip, int len)
{
  long sum=0;
  while(len > 1){
    sum += *(unsigned short*)ip;
    ip+=2;
    if(sum&0x80000000) sum = (sum & 0xffff) + (sum >> 16);
    len-=2;
  }
  if(len)sum+=(unsigned short)*ip;

  while(sum >> 16)sum = (sum & 0xffff) + (sum >> 16);
  return (~sum);
}
      

int valid_ip(char *s)
{
  int i,k,n; 

  n=strlen(s);
  if(n<7 || n>15)return 0;
  
  for(i=0,k=0;i<n;++i){
    if(s[i]=='.')++k;
  }
  if(k!=3)return 0;
  return 1;
}

double gettimeasdouble()
{
    struct timeval tv; 
    gettimeofday(&tv,NULL);
    return (double)tv.tv_sec+tv.tv_usec/1000000.0;
}


void double2tv(double d, struct timeval *tv)
{
  tv->tv_sec=(long)d;
  tv->tv_usec=(long)((d-(long)d)*1000000.0);  
  return;
}

double tv2double(struct timeval *tv)
{
  return (double)tv->tv_sec+tv->tv_usec/1000000.0;
}


char *n2char(unsigned long l)
{
  struct in_addr in;
  in.s_addr=l;
  return inet_ntoa(in);
}

unsigned long char2n(char *s)
{
  return inet_addr(s);
}

int nm2n(unsigned long l)
{
  int i=0;
  for(unsigned long k=0x80000000;k;k>>=1){
    if(l&k) ++i;
  }
  return i; 
}

unsigned long n2nm(int n)
{
  int i;
  unsigned long nm=0;

  for(i=0;i<8 && i+24<n ;++i)
    {
      nm|=(0x80000000>>i);
    }

  for(i=0;i<8 && i+16<n ;++i)
    {
      nm|=(0x00800000>>i);
    }

  for(i=0;i<8 && i+8<n ;++i)
    {
      nm|=(0x00008000>>i);
    }

  for(i=0;i<8 && i<n ;++i)
    {
      nm|=(0x00000080>>i);
    }

  return nm;
}

char *ipnm2char(unsigned long ip, unsigned long nm)
{
  static char buffer[40];
  buffer[0]=0;
  strcpy(buffer,n2char(ip));
  strcat(buffer,"/");
  strcat(buffer,n2char(nm));
  return buffer;
}

int char2ipnm(char *s, unsigned long *ip, unsigned long *nm)
{
  char buffer[128];
  char *p;
  int k=0;
  
  for(p=s;*p;++p)if(*p=='.')++k;

  if(k!=3)return 1;		// error
  
  strcpy(buffer,s);
  p=strchr(buffer,'/'); 
  if(p==NULL){

    *ip=char2n(buffer);
    *nm=0;
    if(*ip&0xff000000)*nm|=0xff000000;
    if(*ip&0x00ff0000)*nm|=0x00ff0000;
    if(*ip&0x0000ff00)*nm|=0x0000ff00;
    if(*ip&0x000000ff)*nm|=0x000000ff;
    return 0;
  }
  *p='\0';++p;
  *ip=char2n(buffer);
  *nm=n2nm(atoi(p));
  return 0;
}



char *dashes(int n)
{
  static char buffer[258];
  int i;
  if(n>256)n=256;
  for(i=0;i<n;++i)buffer[i]='-';
  buffer[i]=0;
  return buffer;
}


char *dashed(char *s, int n)
{
  static char buffer[258];
  int i;
  strcpy(buffer,"-- ");
  strcat(buffer,s);
  strcat(buffer," ");
  for(i=strlen(buffer);i<n;++i)buffer[i]='-';
  buffer[i]=0;
  return buffer;
}

char *dashed(char *s)
{
  return dashed(s,80);
}

char *pkt2char(Packet *p)
{
  static char s[256];

  if(!p){
    return "pkt is null";
  }
  int l=0;

  l+=sprintf(s+l,"%s > ",n2char(PKT_SOURCE(p)));
  l+=sprintf(s+l,"%s ",n2char(PKT_DEST(p)));
  l+=sprintf(s+l,"len:%d dlen:%d id:%u off:%u ttl:%d ",
	     PKT_LENGTH(p),PKT_DATALEN(p),(unsigned short)(PKT_ID(p)),(unsigned short)(PKT_FOFF(p)), PKT_TTL(p));
  if(PKT_MF(p)){l+=sprintf(s+l,"MF ");}
  if(PKT_DF(p)){l+=sprintf(s+l,"DF ");}
  switch(PKT_PROTO(p))
    {
    case 1:
      l+=sprintf(s+l,"icmp ");break;
    case 17:
      l+=sprintf(s+l,"udp ");break;
    default: 
      l+=sprintf(s+l,"proto: %d",PKT_PROTO(p));break;
    }

  l+=sprintf(s+l," tos: %u",PKT_TOS(p));
  
 
  if(PKT_IHL(p)!=5){
    char* cmax=(char*)p+PKT_IHL(p)*4;
    l+=sprintf(s+l," opt: ");
    for(char *c=(char*)p+20; IPOPT_NUMBER(*c)!=IPOPT_END && c<cmax;){

      switch(IPOPT_NUMBER(*c))
        {
        case IPOPT_NOOP: 
	  l+=sprintf(s+l,"nop ");
	  c++;
          break;
        case IPOPT_SEC: 
	  l+=sprintf(s+l,"sec ");
          c+=11; 
	  break;
        case IPOPT_SID:
	  l+=sprintf(s+l,"sid ");
          c+=4; 
	  break;
        case IPOPT_RA:
	  l+=sprintf(s+l,"ra ");
          c+=4; 
	  break;
        default:
	  c+=*(c+1)&0xff;
        }
    }
  }
  return s;
}


unsigned long hostname2n(char *h)
{
  struct hostent *hent;
  hent=gethostbyname(h);
  if(!hent)return 0;
  return *(((unsigned long**)hent->h_addr_list)[0]);
}


int ip_header_check(Packet *p, unsigned short len)
{

  if(len && PKT_LENGTH(p)!=len)return 1;      // packet length
  if(p->version!=4)return 10;	              // ip version
  if(cksum((char*)p,PKT_HDRLEN(p)))return 20; // ip checksum

  return 0;
}


char *proto_id2char(int i)
{
  static char buff[16];
  switch(i)
    {
    case 0: 
      return "IP"; 
    case 1: 
      return "ICMP";
    case 6: 
      return "TCP";
    case 17: 
      return "UDP";
    default:
      sprintf(buff,"%u",i);
      return buff;
    }
}
