#include <stdlib.h>
#include <string.h>
#include "config.H"

#include "vr.H"
#include "tools.H"
#include "api.H"
#include "forwarder.H"
#include "ioc.H"
#include "udp.H"
#include "icmp.H"

UDPstack::UDPstack(VR *v): ProtocolStack()
{
  vr=v;
  memset(udp_ports, 0, sizeof(UDPPORT)*MAX_OPEN_UDPPORTS);
  return;
}

int UDPstack::process(PSPacket *p)
{
  UDPPORT *up;
  up=query_udp_port(UDP_DPORT((UDPPacket*)p->data));
  
  if(!up){
#ifdef DEBUG_UDPSTACK
    fhprintf(vr->cfh, "udpstack: got packet '%s' (unbound)\n",
	     pkt2char(&p->pkt));
#endif
    vr->icmpstack->unreachable(p->dest, &p->pkt, 3);

    return 0;
  }

#ifdef DEBUG_UDPSTACK
  fhprintf(vr->cfh, "udpstack: got packet '%s' (bound)\n",
	     pkt2char(&p->pkt));
#endif

  up->ch->send(p->data+8, p->datalen-8);
    
  return 0;
}

UDPstack::~UDPstack()
{
  return;
}


int  UDPstack::free_udp_port(int p)
{
  int i=p%MAX_OPEN_UDPPORTS;

  while(1){
    if(!udp_ports[i].block)return UDP_NOPORT;
    if(udp_ports[i].busy && udp_ports[i].port_no==p) break;
    i=(i+1)%MAX_OPEN_UDPPORTS;
  }

  udp_ports[i].port_no=0;
  udp_ports[i].busy=0;
  if(!udp_ports[(i+1)%MAX_OPEN_UDPPORTS].block)udp_ports[i].block=0;

  alloc_ports--;
  return 0;
}

int UDPstack::alloc_udp_port(int p, UDP *c)
{
  if(alloc_ports>=MAX_OPEN_UDPPORTS/2) return UDP_NOPORT;
  if(query_udp_port(p))return UDP_BUSY;

  int i=p%MAX_OPEN_UDPPORTS;

  while(1){
    if(!udp_ports[i].busy){
      udp_ports[i].port_no=p;
      udp_ports[i].busy=1;
      udp_ports[i].block=1;
      udp_ports[i].ch=c;
      alloc_ports++;
      return UDP_OKAY;
    }
    i=(i+1)%MAX_OPEN_UDPPORTS;
  }
  return 0; // never reached
}

UDPPORT* UDPstack::query_udp_port(int p)
{

  int i=p%MAX_OPEN_UDPPORTS;
  while(1){
    if(!udp_ports[i].block)return NULL;
    if(udp_ports[i].busy && udp_ports[i].port_no==p) return &udp_ports[i];
    i=(i+1)%MAX_OPEN_UDPPORTS;
  }
  return NULL; // never reached
}


UDP::UDP(VR *v, int p) : IOchannel()
{
  vr=v;
  us=vr->udpstack;
  up=-1;
  if(us->alloc_udp_port(p, this)!=UDP_OKAY){
    up=p;
  }
  up=p;
  return;
}

UDP::~UDP() 
{
  if(up!=-1){
    us->free_udp_port(up);
  }
  return;
}




