#ifndef _vr_h_
#define _vr_h_

#include "config.H"
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <dlfcn.h>

#define IO_WRITE 1         // write possible 
#define IO_READ  2         // read possible
#define TIMER    4         // throw event each timer_time (secs)
#define SINGLE   8         // kill event after one execution

/*
 * Virtual Router container class 
 */
class VR
{
  friend int main(int argc, char **argv);
  friend class Sock;
  friend class UDPSock;
  friend class RawSock;
  friend class LOBenv;
  friend class Shell;
  friend class CoreObject;
  friend class VRsched;
  friend class Routing;
  friend class ICMP;
  friend class UDP;
  friend class UDPstack;
  friend class Interface;
  friend class QCroot;
  friend class VRloadableObject;
  friend class Forwarder;
  friend class Defragmenter;
  friend class API;
  friend class QCO;
  friend class Tunnel_Endpoint;

public:
  VR();
  ~VR();
  double uptime();
  int init();
  int run();
  Interface *get_if(char *id);
  int nof_if(){return num_of_interfaces;}
  Interface *idx_if(int i){return in[i];}
  int add_interface(Interface*);
  int delete_interface(int);
  unsigned long debuglevel;
  VRsched    *vrsched;
  int cfh;			//- the console file handle
				//- all user output should be sent here 


  // functions to set parameters
  int configurebyfile(char *);
  int set_startupfile(char *);
  int set_portnumber(int);
  int set_hostname(char *);
  int set_fifobase(char *);
  int set_lob_path(char *);
  
  //- Configuration values These values should be set before the VR
  //- calls any other constructor
  
  // Shell specific

  int shell_binary;		//- true or false 
				//- shell starts up in binary mode if true  

  int shell_port;		//- port number
				//- if 0 shell listens to stdin/stdout

  char *shell_hostname;		//- name of the host the shell binds to
				//- is filled by gethostname if NULL

  char *shell_startupscript;

  char *shell_password;
  int shell_password_tries;

  // object loadeer

  char *lob_path;		//- path to the precompiled objects
  char *lob_extension;		//- extension of th lob files (.vr)

  // device files

  char *fifo_basename;
  char *softlink_basename;

  // interface parameters

  unsigned long if_bandwidth;
  unsigned long if_bucketsize;

  // defaults for the queueing system

  unsigned long  tbf_bandwidth;
  unsigned long  tbf_bucketsize;
  unsigned short dpt_queuelen;

  // address translation

  int  iptrans_enable;
  unsigned long iptrans_rx_address;
  unsigned long iptrans_rx_netmask;
  unsigned long iptrans_rx_value;
  unsigned long iptrans_rx_mask;	
  
  unsigned long iptrans_tx_address;
  unsigned long iptrans_tx_netmask;
  unsigned long iptrans_tx_value;	
  unsigned long iptrans_tx_mask;

  unsigned short ttl_default;

private:
  double startup_time;
  int num_of_interfaces;	// wie viele interfaces sind definiert
  int highest_interface_idx;	// hoechster interface index
  Interface  *in[MAX_NOF_INTERFACES];
  Forwarder  *pf;
  ICMP       *icmpstack;
  UDPstack   *udpstack;
  Routing    *routing;
  Shell      *shell;
  Tunnel_Endpoint *tunnel_ep;
  
};




#endif











