#include <stdlib.h>   
#include <sys/time.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <sys/stat.h>
 
#include "vr.H"
#include "ioc.H"
#include "tools.H"
#include "lob.H"
#include "vrevent.H"
#include "api.H"
#include "vrsched.H"

void* lob_run(LOBenv *th)
{
  int result=th->lob->run(); 
  if(result==LOB_TERMINATE){
    delete th;
  }
  return NULL;
}

#ifdef DEBUG_EVENTSCHEDULER
FILE *fp_log;
#endif


VRsched::VRsched(VR *m)
{

  for(int i=0;i<256;++i){delay_shall[i]=0; delay_is[i]=0;}

  read_rr=0;
  vr=m;
  event_id=0;
  num_of_events=0;
  events=(VRevent**)malloc(sizeof(VRevent*)*MAX_NUM_OF_EVENTS);
  last_io_start=0;

  num_of_objs=0;
  lws=(LOBenv**)malloc(sizeof(LOBenv*)*MAX_NUM_OF_LOADABLE_OBJECTS);
  memset(lws,0,sizeof(LOBenv*)*MAX_NUM_OF_LOADABLE_OBJECTS);

  act_time=gettimeasdouble();
  //  mincycle=1.0;
  mincycle=0.25;
  event_counter=0;
  events_per_second=0;

  return;
}

VRsched::~VRsched()
{
  for(int i=0;i<num_of_objs;++i)if(lws[i]!=NULL)delete lws[i];
  free(lws);
  free(events);
}


void VRsched::core()
{
  while(1) {
    prepare_select();
    int r=select(fh_max,&rfds, &wfds, &xfds, &sched_tv);
    process_ioevents(r);	// r is true if IO event !!
  }
  return;
}

void VRsched::process_ioevents(int sr)
{
  act_time=gettimeasdouble();
 
  /*
   * okay erst mal alles was nicht IO ist
   */

#ifdef DEBUG_EVENTSCHEDULER
  char logline[256];
  memset(logline,' ',256);
  logline[50]=0;
  for(int i=0;i<10;++i)logline[i*4+3]=':';
#endif

  /* 
   * process forced and timer
   */

  for(int i=0;i<num_of_events;++i){
    VRevent *e=events[i];
    if(e){
				// FORCED events
      if(e->forced){	
	e->o->interrupt(e);
	e->last_call=act_time;
#ifdef DEBUG_EVENTSCHEDULER
	logline[i*4]='F';
#endif
	continue;
      }
				// TIMER events
      if(e->type&TIMER && act_time >= (e->last_call+e->timer_time)){
	e->o->interrupt(e);
	e->last_call=act_time;
#ifdef DEBUG_EVENTSCHEDULER
	logline[i*4]='T';
#endif
	continue;
      }
    }
  }
    
  /*
   * wenn das select ergebis (sr) 0 ist, dann hatten wir kein IO event
   */

  if(sr){			// wenn keine IO events skip write & read
    /*
     * process writes
     */
    for(int i=0;i<num_of_events;++i){
      int k;
      k=i;
      VRevent *e=events[k];
      if(e && e->type&IO_WRITE){
	if(FD_ISSET(e->fh_w, &wfds)){
	  e->hits++;
	  e->thrown=IO_WRITE;
	  e->o->interrupt(e);
	  e->last_call=act_time;
#ifdef DEBUG_EVENTSCHEDULER
	  logline[k*4+2]='W';
#endif
	}
      }
    }
    
    /*
     * process reads
     */
    
    if(read_rr>=num_of_events)read_rr=0;
    int k=read_rr;
    int n=0;
    for(int i=0;i<num_of_events;++i){
      VRevent *e=events[k];
      if(e && e->type&IO_READ){
	if(FD_ISSET(e->fh_r, &rfds)){
	  e->hits++;
	  e->thrown=IO_READ;
	  e->o->interrupt(events[k]);
	  e->last_call=act_time;
	  
#ifdef DEBUG_EVENTSCHEDULER
	  logline[k*4+1]='R';
#endif
	  n++;
	}
      }
      k++;
      if(k>=num_of_events)k=0;
    }
    if(n)read_rr++;
    
    event_counter++;
    
    last_io_start+=num_of_events;
    if(last_io_start>=1024-num_of_events)last_io_start=0;

  } // IO Events

  /*
   * okay, clean ups und logfile schreiben 
   */



#ifdef DEBUG_EVENTSCHEDULER
  if(fp_log==NULL){
    char g[256], h[256];
    gethostname(h,256);
    sprintf(g,"log.%s",h);
    fp_log=fopen(g,"w");
    for(int i=0;i<num_of_events;++i)if(events[i])fprintf(fp_log,"%d:%s\n", i, events[i]->o->id());
      
  }
  fprintf(fp_log,"%f %d %lu %s\n",gettimeasdouble(), read_rr, event_counter, logline); fflush(fp_log);
#endif

  return;
}


void VRsched::prepare_select()
{
  int i;
  act_time=gettimeasdouble();
  double next_time=act_time+mincycle;

  FD_ZERO(&rfds); FD_ZERO(&wfds); FD_ZERO(&xfds);

  for(i=0,fh_max=0;i<num_of_events;++i){
    VRevent *e=events[i];
    if(!e)continue;

    if(e->type&TIMER && !(e->suspended&TIMER)) {
      double t=e->last_call+e->timer_time;
      if(t<next_time)next_time=t;
    }

    if(e->suspended) {
      if(e->suspend_time<=act_time) {
	e->suspended=0;
      }
      else {
	if(e->suspend_time<next_time){
	  next_time=e->suspend_time;
	}
      }
    }

    if(e->type&IO_WRITE && !(e->suspended&IO_WRITE)) {
      if(e->fh_w>fh_max)fh_max=e->fh_w;
      FD_SET(e->fh_w, &wfds);
    }

    if(e->type&IO_READ && !(e->suspended&IO_READ)) {
      if(e->fh_r>fh_max)fh_max=e->fh_r;
      FD_SET(e->fh_r, &rfds);
    }
  }
  ++fh_max;

  if(next_time<act_time)next_time=act_time;
  double2tv(next_time-act_time,&sched_tv);
  return;
}

int VRsched::add_lw(LOBenv *lo)
{
  int i;
  for(i=0;i<MAX_NUM_OF_LOADABLE_OBJECTS;++i){
    if(lws[i]==NULL){
      lws[i]=lo;
      num_of_objs++;
      return i;
    }
  }
  return -1;
}

int VRsched::get_free_loid()
{
  int i;
  for(i=0;i<MAX_NUM_OF_LOADABLE_OBJECTS;++i){
    if(lws[i]==NULL){
      return i;
    }
  }
  return -1;
}

int VRsched::del_lw(int i)
{
  if(lws[i]){
    lws[i]=NULL;
    num_of_objs--;
    return 0;
  }
  return -1;
}

API* LOBenv::create_APIchannel()
{
 return new API(vr);
}


LOBenv* VRsched::load_object(API *calling_api, char *name, int argc, char **argv, int *error)
{
  LOBenv *env=new LOBenv(vr);

  env->calling_api=calling_api;

  strcpy(env->name,name);
  env->argc=argc;
  for(int i=0;i<argc;++i){
    env->argv[i]=strdup(argv[i]);
  }
  
  if(num_of_objs >= MAX_NUM_OF_LOADABLE_OBJECTS){
    delete env;
    *error=LOB_LOAD_EXEED;
    return NULL;
  }
  
#ifdef DEBUG_MODULELOADER
  fhprintf(vr->cfh,"--- DEBUG_MODULELOADER ---\n");
  fhputs("dlopen starts\n",vr->cfh);
#endif
  
  if(strchr(env->name,'/')==NULL && strchr(env->name,'.')==NULL){
    sprintf(env->filename,"%s/%s%s",vr->lob_path,env->name,vr->lob_extension);
  } else {
    strcpy(env->filename,name);
  }

  /*
   * getting file information
   */

  if(stat(env->filename, &env->statbuff)){
#ifdef DEBUG_MODULELOADER
    fhprintf(vr->cfh,"cannot stat file: %s\n",env->filename);
#endif
    delete env;
    *error=LOB_LOAD_OPEN;
    return NULL;
  }
 
  /*
   * loading object
   */ 

  env->handle=dlopen(env->filename, RTLD_GLOBAL|RTLD_NOW);
  if(!env->handle){
#ifdef DEBUG_MODULELOADER
    fhprintf(vr->cfh,"Error at dlopen: %s\n",dlerror());
#endif
    delete env;
    *error=LOB_LOAD_OPEN;
    return NULL;
  }
  
  /*
   * Symbol table lookup
   */

  env->lob_start=(LOB*(*)(LOBenv*))dlsym(env->handle,"lob_start");
  if(env->lob_start==NULL){
#ifdef DEBUG_MODULELOADER
    fhprintf(vr->cfh,"Error at symbol lookup\n",dlerror());
#endif
    delete env;
    *error=LOB_LOAD_SYM;
    return NULL;
  }


  /*
   * the LOB constructor
   */
  env->in_constructor=1;
  env->lob=env->lob_start(env);	// this calls the constructor of the lob 
  env->in_constructor=0;
  env->size=sizeof(*env->lob);
  env->loadtime=gettimeasdouble();
  
  /*
   * the various run modes
   */

  if(env->mode&LOB_MODE_THREAD){ // THREAD
#ifdef DEBUG_MODULELOADER
    fhprintf(vr->cfh,"before thread creation %s\n",env->filename);
#endif
    pthread_create(&env->thread, NULL, (void * (*)(void*))lob_run, (void*)env);
#ifdef DEBUG_MODULELOADER
    fhprintf(vr->cfh,"after thread creation %s\n",env->name);
#endif
    *error=LOB_LOAD_OKAY;
    return env;
  }
  
  if(env->mode&LOB_MODE_RUN){	// RUN
#ifdef DEBUG_MODULELOADER
    fhprintf(vr->cfh,"before exec of lob %s\n",env->filename);
#endif
    lob_run(env);
#ifdef DEBUG_MODULELOADER
    fhprintf(vr->cfh,"after exec of lob %s\n",env->name);
#endif
    *error=LOB_LOAD_OKAY;
    return NULL;
  }
  
  if(env->mode&LOB_MODE_STAY){	// STAY
#ifdef DEBUG_MODULELOADER
    fhprintf(vr->cfh,"lob stays in memory\n");
#endif
    *error=LOB_LOAD_OKAY;
    return env;
  }

  // run wurde nicht aufgerufen und object ist obsolete

  delete env;
  *error=LOB_LOAD_OKAY;
#ifdef DEBUG_MODULELOADER
  fhprintf(vr->cfh,"lob finished and removed\n");
#endif
  return NULL;
}
 

int VRsched::notify(VRevent *e)
{
  int i;

#ifdef DEBUG_EVENTSYSTEM
  puts("vrsched: event notify"); 
#endif
  e->event_id=event_id++;
  e->thrown=0;
      
  for(i=0;i<num_of_events;++i){
    if(events[i]==NULL){
      events[i] = e;
      goto notify_return;
    }
  }
  events[num_of_events++]=e;
  
 notify_return:

  //  mix_up_ev_idx();
  
  return 0;
}

int VRsched::release(unsigned long id)
{
  int i;

  for(i=0;i<num_of_events;++i){
    if(events[i]->event_id==id){
      events[i]=NULL; 
      if(i+1==num_of_events)num_of_events--;
      mix_up_ev_idx();
     return 0;
    }
  }
  return 1;
}


void VRsched::mix_up_ev_idx()
{
  int i,k;
  
  for(k=0;k<1024;k+=num_of_events){
    for(i=0;i<num_of_events;++i){
      ev_idx[i+k]=i;
    }
    
    for(i=0;i<num_of_events;++i){
      int r=(int)((double)num_of_events*rand()/RAND_MAX);
      int s=ev_idx[i+k];
      ev_idx[i+k]=ev_idx[r+k];
      ev_idx[r+k]=s;
    }
  }
  //  for(k=0;k<1024;++k){
  //  printf("%d %d\n",k, ev_idx[k]);
  //  }
  return;
}



LOBenv::LOBenv(VR *v)
{
#ifdef DEBUG_MODULELOADER
  puts("--- DEBUG_MODULELOADER ---\n");
  puts("LOBenv constructor called"); // no console fh !!!
#endif

  vr=v;
  cfh=&(vr->cfh);
  argv=(char**)malloc(64*sizeof(char*));
  memset(argv,0,sizeof(char*)*64);
  memset(name,0,32);
  memset(filename,0,256);
  argc=0;
  loid=vr->vrsched->add_lw(this);
  lob=NULL;
  handle=NULL;
  mode=0;
  in_constructor=0;
  return;
}

LOBenv::LOBenv(LOBenv *e)
{
  in_constructor=e->in_constructor;
  vr=e->vr;
  cfh=e->cfh;
#ifdef DEBUG_MODULELOADER
  fhputs("--- DEBUG_MODULELOADER ---\n",*cfh);
  fhputs("LOBenv copyconstructor called\n",*cfh);
#endif
  argv=(char**)malloc(64*sizeof(char*));
  memset(argv,0,sizeof(char*)*64);
  memcpy(name,e->name,32);
  memcpy(filename,e->filename,256);
  argc=e->argc;
  for(int i=0;i<argc;++i){
    if(e->argv[i])argv[i]=strdup(e->argv[i]);
  }
  loid=e->loid;
  lob=e->lob;
  lob_start=e->lob_start;
  handle=e->handle;
  mode=e->mode;
  return;
}

LOBenv::~LOBenv()
{
#ifdef DEBUG_MODULELOADER
  fhputs("--- DEBUG_MODULELOADER ---\n",*cfh);
  fhputs("LOBenv destructor called\n",*cfh);
#endif
  vr->vrsched->del_lw(loid);
  for(int i=0;i<64;++i)if(argv[i])free(argv[i]);
  if(argv)free(argv);
  if(lob)delete lob; 
  //  if(handle) dlclose(handle);
  return;
}

VRRB* LOBenv::constructor_api_access(VRCB *c)
{
  if(in_constructor){
#ifdef DEBUG_MODULELOADER
    fhputs("--- DEBUG_MODULELOADER ---\n",vr->cfh);
    fhputs("constructor_api_access in constructor mode\n",vr->cfh);
#endif
    return calling_api->process_VRCB(c);
  }

#ifdef DEBUG_MODULELOADER
  fhputs("--- DEBUG_MODULELOADER ---\n",vr->cfh);
  fhputs("constructor_api_access not in constructor mode\n",vr->cfh);
#endif

  return NULL;
}


VRRB* LOBenv::api_direct(VRCB *c)
{
  c->handle=calling_api->get_handle(); 
  return constructor_api_access(c);
}


void LOBenv::finish()
{
  delete this;
}


VRRB* VRsched::APIconfigure(API *calling_api, VRCB *in)
{
  struct dummy0 : public VRCB
  {
    unsigned short lob_command;
  };

  switch(((struct dummy0*)in)->lob_command)
    {
    case 1: {	// add LOB
      struct dummy1: public VRCB
      {
	unsigned short lob_command;	
	unsigned short flags;	
	char name[32];
	char command_line[1];
      } *cb;
      
      struct dummy2: public VRRB
      {
	struct LOB_INFO_BLOCK m;
      } *rb;
      
      cb=(struct dummy1*)in;
      
      rb=(struct dummy2*)malloc(sizeof(struct dummy2));
      
      /*
       *  generate Environment object
       */
   
      char *argbuff[256];
      char *p=cb->command_line;	
      int z=0;
      if(*p)argbuff[z++]=p;p++;
      for(unsigned int i=sizeof(struct dummy1)-1;i<cb->len;++i){
	if(*p==0 && *(p+1)){
	  argbuff[z++]=p+1;
	}
	++p;
      }
      int error;
      	
      LOBenv* lw=load_object(calling_api, cb->name, z, argbuff, &error);

      if(!lw){
	strcpy(rb->m.name,cb->name);
	rb->m.lob_id=0xffff;
	rb->m.mode=error;
      } else {
	strcpy(rb->m.name,lw->name);
	rb->m.lob_id=lw->loid;
	rb->m.mode=lw->mode;
      }
      rb->handle=cb->handle;
      rb->len=sizeof(struct dummy2);

      return (VRRB*)rb;
    }
    case 10: {
      struct dummy0 : public VRRB {
	LOB_INFO_BLOCK ibl;
      } *rb;
      
      rb=(struct dummy0*)malloc(sizeof(VRRB)+sizeof(LOB_INFO_BLOCK)*num_of_objs);
      rb->handle=in->handle;
      
      LOB_INFO_BLOCK *b=&(rb->ibl);

      int j=0;
      for(int i=0;i<MAX_NUM_OF_LOADABLE_OBJECTS;++i){
	if(lws[i]){
	  b[j].lob_id =lws[i]->loid;
	  b[j].mode=lws[i]->mode;
	  b[j].size=lws[i]->statbuff.st_size;
	  b[j].uptime=gettimeasdouble()-lws[i]->loadtime;
	  strncpy(b[j].name, lws[i]->name,32);
	 
	  ++j;
	}
      }
      rb->len=sizeof(VRRB)+sizeof(LOB_INFO_BLOCK)*j;
      return rb;
    }
    }
  return NULL;
}

