#ifndef vrsched_h_
#define vrsched_h_

#include <sys/time.h>
#include <sys/stat.h>
#include <pthread.h>

class VRevent;
class VR;
class LOB;
class LOBenv;
class VRRB;
class VRCB;
class API;

class VRschedObj
{
public:
  virtual char *id() = 0;
  virtual int interrupt(VRevent*) = 0 ;
}; 

class VRsched
{
  friend int main(int,char**);
  friend class API;
  friend class LOBenv;
  friend void* lob_run(LOBenv*);

public:
  VRsched(VR *);
  ~VRsched();
  void core();
  int notify(VRevent *);
  int release(unsigned long id);
  VRRB* APIconfigure(API *, VRCB*);

private:
  VR *vr;

  // loadable object stuff

  int get_free_loid();
  LOBenv **lws;
  int num_of_objs;
  LOBenv *load_object(API *a, char *name, int argc, char **argv, int *error);
  int add_lw(LOBenv *);
  int del_lw(int);


  // event handling
  
  int ev_idx[1024];
  void mix_up_ev_idx();
  void prepare_select();
  void process_ioevents(int);

  fd_set rfds;
  fd_set wfds;
  fd_set xfds;
  struct timeval sched_tv; 
  int fh_max;
  int read_rr;
  int last_io_start;
  unsigned long event_id;
  int num_of_events;
  VRevent **events;
  double act_time;
  double mincycle;
  double events_per_cycle;
  double events_per_second;
  unsigned long event_counter;
  int delay_shall[256];
  int delay_is[256];
 
};


class LOBenv
{
  friend class VRsched;
  friend void* lob_run(LOBenv *);
  
public:
  LOBenv(VR*);
  LOBenv(LOBenv*);
  ~LOBenv();
  int *cfh;
  char **argv;
  int argc;
  char name[32];		// some id for the lob
  
  int set_mode(int i){mode=i; return i;}
  int get_loid(){return loid;}
  VRRB *api_direct(VRCB *c);
  API *create_APIchannel();
  int EventNotify(VRevent *e){return vr->vrsched->notify(e);}
  int EventRelease(unsigned long id){return vr->vrsched->release(id);}
  void finish();

private:
  int in_constructor;
  VRRB* constructor_api_access(VRCB*);
  LOB * (* lob_start)(LOBenv*);
  VR *vr;			//- the reference to the VR 
  LOB *lob;			//- the reference to the LOB
  void *handle;			//- the dl handle of the LOB 
  int loid;			//- the loadable object id 
  char filename[256];
  pthread_t thread;	
  double loadtime;
  unsigned long size;
  unsigned short mode;
  struct stat statbuff;
  API *calling_api;
};




#endif
